from aiohttp import web
import datetime
from datetime import timezone, datetime

from sqlalchemy import desc, asc
from sqlalchemy.orm import Session
from sqlalchemy.exc import SQLAlchemyError

from joule.models.annotation import Annotation, from_json
from joule.models.data_stream import DataStream
from joule.models import folder
from joule.errors import ApiError
from joule.constants import ApiErrorMessages
from joule import app_keys
from joule.utilities import datetime_to_timestamp

async def info(request):
    db: Session = request.app[app_keys.db]

    if 'stream_id' in request.query:
        my_stream = db.get(DataStream,request.query["stream_id"])
    elif 'stream_path' in request.query:
        my_stream = folder.find_stream_by_path(request.query['stream_path'], db, stream_type=DataStream)
    else:
        raise web.HTTPBadRequest(reason=ApiErrorMessages.specify_id_or_path)
    if my_stream is None:
        raise web.HTTPNotFound(reason=ApiErrorMessages.stream_does_not_exist)


    count = db.query(Annotation.start).filter_by(stream_id=my_stream.id).count()
    if count > 0:
        start = db.query(Annotation.start)\
            .filter_by(stream_id=my_stream.id)\
            .order_by(asc(Annotation.start))\
            .limit(1).scalar()
        start_ts = datetime_to_timestamp(start)
        end = db.query(Annotation.start)\
            .filter_by(stream_id=my_stream.id)\
            .order_by(desc(Annotation.start))\
            .limit(1).scalar()
        end_ts = datetime_to_timestamp(end)

    else:
        start_ts = None
        end_ts = None

    return web.json_response({
        'start': start_ts,
        'end': end_ts,
        'count': count
    })


async def index(request):
    db: Session = request.app[app_keys.db]
    # specify stream_ids as array
    # optionally specify start and end
    # parse time bounds
    start, end = _parse_time_bounds(request)
    if (('stream_id' not in request.query) and
            ('stream_path' not in request.query)):
        return web.Response(text="must specify at least one stream_id or stream_path", status=400)

    response = []
    stream_ids = []
    if "stream_id" in request.query:
        stream_ids = request.query.getall("stream_id")
    if "stream_path" in request.query:
        for path in request.query.getall("stream_path"):
            stream = folder.find_stream_by_path(path, db, stream_type=DataStream)
            if stream is None:
                raise web.HTTPNotFound(reason=ApiErrorMessages.stream_does_not_exist)
            stream_ids.append(stream.id)

    for stream_id in stream_ids:
        annotations = db.query(Annotation).filter_by(stream_id=stream_id)
        if start is not None:
            annotations = annotations.filter(Annotation.start >= start)
        if end is not None:
            annotations = annotations.filter(Annotation.start <= end)
        response += [a.to_json() for a in annotations]

    return web.json_response(response)


async def update(request):
    db: Session = request.app[app_keys.db]
    if request.content_type != 'application/json':
        raise web.HTTPBadRequest(reason='content-type must be application/json')
    body = await request.json()

    if 'id' in body:
        my_annotation = db.get(Annotation, body["id"])
    else:
        raise web.HTTPBadRequest(reason="specify an id")
    if my_annotation is None:
        raise web.HTTPNotFound(reason="annotation does not exist")
    my_annotation.update_attributes(body) # cannot change timestamps
    if my_annotation.title is None or my_annotation.title == '':
        db.rollback()
        raise web.HTTPBadRequest(reason="annotation title is reqiured")
    db.commit()
    return web.json_response(my_annotation.to_json())


async def create(request):
    db: Session = request.app[app_keys.db]
    if request.content_type != 'application/json':
        raise web.HTTPBadRequest(reason='content-type must be application/json')
    body = await request.json()

    if 'stream_id' in body:
        my_stream = db.get(DataStream,body["stream_id"])
    elif 'stream_path' in body:
        my_stream = folder.find_stream_by_path(body['stream_path'], db, stream_type=DataStream)
    else:
        raise web.HTTPBadRequest(reason="specify a stream_id")
    if my_stream is None:
        raise web.HTTPNotFound(reason=ApiErrorMessages.stream_does_not_exist)

    try:
        my_annotation = from_json(body)
        if my_annotation.title is None or my_annotation.title == '':
            raise ApiError("title is reqiured")
        if my_annotation.end is not None and (my_annotation.start >= my_annotation.end):
            raise web.HTTPBadRequest(reason=ApiErrorMessages.start_must_be_before_end)
    except ApiError as e:
        raise web.HTTPBadRequest(reason=str(e))

    my_annotation.stream = my_stream
    db.add(my_annotation)
    db.commit()
    return web.json_response(my_annotation.to_json())


async def delete(request):
    db: Session = request.app[app_keys.db]
    if 'id' in request.query:
        my_annotation = db.get(Annotation,request.query["id"])
    else:
        raise web.HTTPBadRequest(reason="specify an id")
    if my_annotation is None:
        raise web.HTTPNotFound(reason="annotation does not exist")
    db.delete(my_annotation)
    db.commit()
    return web.Response(text="ok")


async def delete_all(request):
    db: Session = request.app[app_keys.db]

    if "stream_id" in request.query:
        stream_id = request.query["stream_id"]
        my_stream = db.get(DataStream,stream_id)
    elif "stream_path" in request.query:
        path = request.query["stream_path"]
        my_stream = folder.find_stream_by_path(path, db, stream_type=DataStream)
    else:
        raise web.HTTPBadRequest(reason=ApiErrorMessages.specify_id_or_path)

    start, end = _parse_time_bounds(request)

    if my_stream is None:
        raise web.HTTPNotFound(reason=ApiErrorMessages.stream_does_not_exist)

    annotations = db.query(Annotation).filter_by(stream_id=my_stream.id)
    if start is not None:
        annotations = annotations.filter(Annotation.start >= start)
    if end is not None:
        annotations = annotations.filter(Annotation.start <= end)

    for annotation in annotations:
        db.delete(annotation)
    db.commit()
    return web.Response(text="ok")

def _parse_time_bounds(request):
    start = None
    end = None
    try:
        if 'start' in request.query:
            ts = int(request.query['start'])
            start = datetime.fromtimestamp(ts/1e6, timezone.utc)
        if 'end' in request.query:
            ts = int(request.query['end'])
            end = datetime.fromtimestamp(ts/1e6, timezone.utc)
    except ValueError:
        raise web.HTTPBadRequest(reason="[start] and [end] must be microsecond utc timestamps")
    return start, end


from aiohttp import web

from joule.models.supervisor import Supervisor
from joule import app_keys
from joule.constants import ApiErrorMessages

async def index(request):
    supervisor: Supervisor = request.app[app_keys.supervisor]
    resp = []
    resp = [p.to_json() for p in supervisor.proxies]
    return web.json_response(data=resp)


async def info(request):
    supervisor: Supervisor = request.app[app_keys.supervisor]
    if 'name' in request.query:
        proxy = [p for p in supervisor.proxies if p.name == request.query['name']]
    elif 'id' in request.query:
        proxy = [p for p in supervisor.proxies if p.uuid == int(request.query['id'])]
    else:
        raise web.HTTPBadRequest(reason=ApiErrorMessages.specify_id_or_path)
    if len(proxy) == 0:
        raise web.HTTPNotFound(reason="proxy does not exist")
    proxy = proxy[0]
    return web.json_response(proxy.to_json())



#!/usr/bin/env -S python3 -u

"""
Run the configuration in follower
"""

import os
import sys
import subprocess
import shutil
from helpers import wait_for_joule_host
SOURCE_DIR = "/joule"
MODULE_SCRIPT_DIR = "/joule/tests/e2e/module_scripts"
JOULE_CONF_DIR = "/joule/tests/e2e/follower"
JOULE_CMD = "coverage run --rcfile=/joule/.coveragerc -m joule.cli".split(" ")



def main():
    subprocess.run(("joule admin erase --yes").split(" "))

    # make module scripts available
    os.symlink(MODULE_SCRIPT_DIR, "/module_scripts")

    # get API permissions
    os.environ["LOGNAME"] = "e2e"
    os.environ["JOULE_USER_CONFIG_DIR"] = "/tmp/joule_user"
    
    # copy module configurations (no stream configs)
    shutil.rmtree("/etc/joule/module_configs")
    shutil.copytree(os.path.join(JOULE_CONF_DIR, "module_configs"), "/etc/joule/module_configs")

    jouled = subprocess.Popen("jouled",
                              stdout=sys.stdout,
                              stderr=sys.stderr,
                              universal_newlines=True)
    
    ## The follower tests out the built-in HTTP server
    wait_for_joule_host('node2.joule', port=8080)

    subprocess.run(JOULE_CMD+"admin authorize".split(" "))

    # follow node1.joule
    wait_for_joule_host('node1.joule')
    result = subprocess.run("joule master add joule http://node1.joule/joule".split(" "))  # ,
    if result.returncode != 0:
        print("ERROR: failed to add master node")
        jouled.send_signal(subprocess.signal.SIGINT)

    # this will just hang, node1 exits and terminates this container
    stdout, stderr = jouled.communicate()
    print("---- jouled stdout ----")
    print(stdout)
    print("---- jouled stderr ----")
    print(stderr)
    return 0


if __name__ == "__main__":
    exit(main())



from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web
from sqlalchemy.orm import Session
import asyncio
from joule.models import folder, DataStream, Folder, Element, StreamInfo, EventStream
from joule.api.event_stream import EventStream as ApiEventStream
import joule.controllers
from tests.controllers.helpers import create_db, MockStore, MockEventStore
from joule import app_keys
from joule.constants import EndPoints
import testing.postgresql
psql_key = web.AppKey("psql", testing.postgresql.Postgresql)

class TestFolderController(AioHTTPTestCase):

    async def tearDownAsync(self):
        self.app[app_keys.db].close()
        self.app[psql_key].stop()
        await self.client.close()

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        # this takes a while, adjust the expected coroutine execution time
        loop = asyncio.get_running_loop()
        loop.slow_callback_duration = 2.0
        # create two event streams to try recursive dele\ operations
        stream1 = ApiEventStream(name="test",description="test stream",keep_us=10e6, 
                            event_fields={"field1": "string",
                                        "field2": 'category:["1","2","3"]'})
        stream2 = ApiEventStream(name="test2",description="test stream2",keep_us=10e6,
                            event_fields={"field1": "string",
                                        "field2": 'category:["1","2","3"]'})
        # place the events under /top/leaf to test delete operations
        app[app_keys.db], app[psql_key] = create_db([
            "/other/middle/stream3:int16[val1, val2]",
            "/top/leaf/stream1:float32[x, y, z]",
            "/top/middle/leaf/stream2:int16[val1, val2]"],
            [stream1,stream2],"/top/leaf")
        app[app_keys.data_store] = MockStore()
        app[app_keys.event_store] = MockEventStore()
        return app

    async def test_stream_list(self):
        db: Session = self.app[app_keys.db]
        my_stream: DataStream = db.query(DataStream).filter_by(name="stream1").one()
        store: MockStore = self.app[app_keys.data_store]
        mock_info = StreamInfo(start=0, end=100, rows=200)
        store.set_info(my_stream, mock_info)

        resp = await self.client.request("GET", EndPoints.folders)
        actual = await resp.json()
        # list of active streams should be in the response
        self.assertTrue("active_data_streams" in actual)
        # remove active_streams from the response, so we can compare it with the database copy
        del actual['active_data_streams']
        # basic check to see if JSON response matches database structure
        expected = folder.root(db).to_json()
        self.assertEqual(actual, expected)
        # test to see if data-info flag works
        resp = await self.client.request("GET", EndPoints.folders, params={"data-info": ""})
        actual = await resp.json()
        expected = folder.root(db).to_json(data_stream_info={my_stream.id: mock_info})
        # remove active_streams from the response, so we can compare it with the database copy
        del actual['active_data_streams']
        self.assertEqual(actual, expected)


    async def test_folder_info(self):
        # query by path
        params = {
            "path": "/top/leaf"
        }
        resp = await self.client.get(EndPoints.folder, params=params)
        self.assertEqual(resp.status, 200)
        json = await resp.json()
        self.assertEqual(json["name"], "leaf")
        self.assertEqual(len(json["streams"]), 1)
        folder_id = json["id"]
        # query by id
        resp = await self.client.get(EndPoints.folder, params={"id": folder_id})
        self.assertEqual(resp.status, 200)
        json = await resp.json()
        self.assertEqual(json["name"], "leaf")
        self.assertEqual(len(json["streams"]), 1)


    async def test_folder_move_by_path(self):
        db: Session = self.app[app_keys.db]
        # move stream1 into folder3
        payload = {
            "src_path": "/top/leaf",
            "dest_path": "/other"
        }
        source_parent_created_at = folder.find("/top", db).updated_at
        dest_parent_created_at = folder.find("/other", db).updated_at
        other_folder_created_at = folder.find("/other/middle", db).updated_at
        resp = await self.client.put(EndPoints.folder_move, json=payload)
        self.assertEqual(resp.status, 200)
        f = folder.find("/other/leaf", db)
        self.assertEqual(f.data_streams[0].name, "stream1")
        self.assertIsNone(folder.find("/top/leaf", db))
        # make sure the parent timestamps are updated
        self.assertGreater(folder.find("/top", db).updated_at, source_parent_created_at)
        self.assertGreater(folder.find("/other", db).updated_at, dest_parent_created_at)
        # other timestamps should not be updated
        self.assertEqual(folder.find("/other/middle", db).updated_at, other_folder_created_at)

    async def test_folder_move_by_id(self):
        db: Session = self.app[app_keys.db]
        # move stream1 into folder3
        dest_folder = folder.find("/top/middle/leaf", db)
        src_folder = folder.find("/top/leaf", db)
        payload = {
            "src_id": src_folder.id,
            "dest_id": dest_folder.id
        }
        resp = await self.client.put(EndPoints.folder_move, json=payload)
        self.assertEqual(resp.status, 200)
        self.assertEqual(src_folder.id,
                         folder.find("/top/middle/leaf/leaf", db).id)
        self.assertIsNone(folder.find("/top/leaf", db))


    async def test_folder_delete_by_path(self):
        db: Session = self.app[app_keys.db]
        folder.find("/top/leaf", db)
        payload = {'path': "/top/leaf"}
        # there are 2 event streams associated with this folder
        self.assertEqual(2, db.query(EventStream).count())
        resp = await self.client.delete(EndPoints.folder, params=payload)
        self.assertEqual(resp.status, 200)

        self.assertIsNone(folder.find("/top/leaf", db))
        # deletes the streams
        self.assertIsNone(folder.find_stream_by_path("/top/leaf/stream1", db))
        # deletes the event streams
        self.assertIsNone(folder.find_stream_by_path("/top/leaf/test", db))
        self.assertEqual(0, db.query(EventStream).count())
        # keeps the parent folders
        self.assertIsNotNone(folder.find("/top", db))


    async def test_folder_delete_by_id(self):
        db: Session = self.app[app_keys.db]
        f_count = db.query(Folder).count()
        f = folder.find("/an/empty/folder", db, create=True)
        parent_updated_at = f.parent.updated_at
        top_parent_updated_at = f.parent.parent.updated_at
        self.assertEqual(db.query(Folder).count(), f_count + 3)
        payload = {'id': f.id}
        resp = await self.client.delete(EndPoints.folder, params=payload)
        self.assertEqual(resp.status, 200)

        self.assertIsNone(folder.find("/an/empty/folder", db))
        # keeps the parent folders
        self.assertEqual(f_count + 2, db.query(Folder).count())
        self.assertIsNotNone(folder.find("/an/empty", db))
        # parent folders should have updated timestamps
        self.assertGreater(f.parent.updated_at, parent_updated_at)
        self.assertGreater(f.parent.parent.updated_at, top_parent_updated_at)

    async def test_folder_recursive_delete(self):
        db: Session = self.app[app_keys.db]
        f = folder.find("/top", db)
        payload = {'path': "/top", 'recursive': "1"}
        resp = await self.client.delete(EndPoints.folder, params=payload)
        self.assertEqual(resp.status, 200)
        # this is the top folder so everything under it should be gone
        # there are three other folders: /, /other, and /other/middle
        self.assertEqual(3, db.query(Folder).count())
        # there is one other stream: /other/middle/stream3:int16[val1, val2]
        self.assertEqual(1, db.query(DataStream).count())
        self.assertEqual(2, db.query(Element).count())
        # the event streams are gone
        self.assertEqual(0, db.query(EventStream).count())

    async def test_folder_update(self):
        db: Session = self.app[app_keys.db]
        my_folder = folder.find("/top/middle/leaf", db)
        other_folder = folder.find("/top/other", db, create=True)
        created_at = my_folder.updated_at
        middle_parent_created_at = my_folder.parent.updated_at
        top_parent_created_at = my_folder.parent.parent.updated_at
        other_folder_created_at = other_folder.updated_at
        # change the stream name
        payload = {
            "id": my_folder.id,
            "folder": {"name": "new name", "description": "new description"}
        }
        resp = await self.client.put(EndPoints.folder, json=payload)
        self.assertEqual(200, resp.status)
        my_folder: DataStream = db.get(Folder,my_folder.id)
        self.assertEqual("new name", my_folder.name)
        self.assertEqual("new description", my_folder.description)
        # make sure updated timestamps are more recent than created timestamps
        self.assertGreater(my_folder.updated_at, created_at)
        self.assertGreater(my_folder.parent.updated_at, middle_parent_created_at)
        self.assertGreater(my_folder.parent.parent.updated_at, top_parent_created_at)
        # make sure other folder is not updated
        self.assertEqual(other_folder_created_at, other_folder.updated_at)
        # make sure the JSON response is correct
        json = await resp.json()
        self.assertEqual("new name", json["name"])
        self.assertEqual(my_folder.updated_at.isoformat(), json["updated_at"])



import unittest
from unittest import mock
from joule.errors import ConfigurationError
from joule.models import data_stream  # for helper functions
from tests import helpers
from datetime import datetime, timezone
from joule.models import (DataStream, Element, data_stream)
from joule.models.data_stream import from_json as data_stream_from_json


class TestStreamErrors(unittest.TestCase):

    def setUp(self):
        self.base_config = helpers.parse_configs(
            """[Main]
                 name = test
                 description = test_description                 
                 path = /some/path/to/data
                 datatype = float32
                 keep = 1w
                 decimate = yes
               [Element1]
                 name = e1
            """)

    def test_errors_on_missing_main(self):
        self.base_config.remove_section("Main")
        with self.assertRaisesRegex(ConfigurationError, 'Main'):
            data_stream.from_config(self.base_config)

    def test_errors_on_bad_name(self):
        bad_names = ["", "has/slash", "/other"]
        self.evaluate_bad_values("name", bad_names)

    def test_errors_on_missing_info(self):
        self.base_config.remove_option("Main", "name")
        with self.assertRaisesRegex(ConfigurationError, 'name'):
            data_stream.from_config(self.base_config)

    def test_errors_on_bad_keep(self):
        """keep is # and timeunit (eg 1w, 30h, 2y) or False"""
        bad_keeps = ["0w", "3", "w", "something random", "-2h"]
        self.evaluate_bad_values("keep", bad_keeps)

    def test_errors_on_bad_datatype(self):
        bad_datatypes = ["", "intz", "0"]
        self.evaluate_bad_values("datatype", bad_datatypes)

    def test_errors_on_missing_elements_sections(self):
        """Must have at least one element"""
        self.base_config.remove_section("Element1")
        with self.assertRaises(ConfigurationError):
            data_stream.from_config(self.base_config)

    def test_errors_on_update_from_incompatible_stream(self):
        """Streams must have the same number of elements"""
        my_stream = data_stream.DataStream(name="test",
                                           description="test_description",
                                           datatype=DataStream.DATATYPE.FLOAT32,
                                           keep_us=DataStream.KEEP_ALL,
                                           decimate=True,
                                           updated_at=datetime.now(timezone.utc))
        my_stream.elements.append(Element(name="e1", index=0,
                                          display_type=Element.DISPLAYTYPE.CONTINUOUS))
        other_stream = data_stream_from_json(my_stream.to_json())
        other_stream.elements.append(Element(name="e2", index=1,
                                             display_type=Element.DISPLAYTYPE.CONTINUOUS))
        with self.assertRaisesRegex(ConfigurationError, "elements"):
            my_stream.merge_configs(other_stream)

    @mock.patch('joule.models.element.from_config')
    def test_errors_on_invalid_elements(self, from_config: mock.Mock):
        from_config.side_effect = ConfigurationError()
        with self.assertRaisesRegex(ConfigurationError, 'element'):
            data_stream.from_config(self.base_config)

    def evaluate_bad_values(self, setting_name, bad_settings, section="Main"):
        for setting in bad_settings:
            with self.subTest(setting=setting):
                self.base_config[section][setting_name] = setting
                with self.assertRaisesRegex(ConfigurationError, setting_name):
                    data_stream.from_config(self.base_config)

    def evaluate_good_values(self, setting_name, settings, section="Main"):
        for setting in settings:
            with self.subTest(setting=setting):
                self.base_config[section][setting_name] = setting
                data_stream.from_config(self.base_config)



import sqlalchemy
import ipaddress
from urllib.parse import urlparse

from .session import BaseSession
from joule.utilities import connection_info
from joule.constants import EndPoints


async def db_connect(session: BaseSession) -> sqlalchemy.engine.Engine:
    conn_info = await db_connection_info(session)
    return sqlalchemy.create_engine(conn_info.to_dsn())


async def db_connection_info(session: BaseSession) -> connection_info.ConnectionInfo:
    resp = await session.get(EndPoints.db_connection)
    conn_info = connection_info.from_json(resp)
    if is_localhost(conn_info.host):
        # replace the host with the session host
        url = urlparse(session.url)
        conn_info.host = url.hostname
    return conn_info

def is_localhost(hostname) -> bool:
    if hostname == 'localhost':
        return True
    try:
        ip = ipaddress.ip_address(hostname)
        if ip.is_loopback:
            return True
    except ValueError:
        # If it's not an IP address, it might be a domain name, which is not localhost
        return False
    # It is an IP address but not a loopback address
    return False



from click.testing import CliRunner
import os
import asyncio

from aiohttp.test_utils import unused_port
import warnings
import json
import copy

from ..fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main

MODULE_INFO = os.path.join(os.path.dirname(__file__), 'module.json')
warnings.simplefilter('always')


class TestModuleInfo(FakeJouleTestCase):

    def test_shows_module_info(self):
        server = FakeJoule()
        with open(MODULE_INFO, 'r') as f:
            server.response = f.read()
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['module', 'info', 'ModuleName'])
        self.assertEqual(result.exit_code, 0)
        # make sure stream inputs and outputs are listed
        self.assertIn("/folder_1/random", result.output)
        self.assertIn("/folder_1/filtered", result.output)
        # check for name and description
        self.assertIn('ModuleName', result.output)
        self.assertIn('the module description', result.output)
        self.stop_server()

    def test_handles_different_module_configurations(self):
        server = FakeJoule()
        with open(MODULE_INFO, 'r') as f:
            orig_model_data = json.loads(f.read())
        module1 = copy.deepcopy(orig_model_data)
        module1['inputs'] = []
        module2 = copy.deepcopy(orig_model_data)
        module2['outputs'] = []
        for module in [module1, module2]:

            server.response = json.dumps(module)

            self.start_server(server)
            runner = CliRunner()
            result = runner.invoke(main, ['module', 'info', 'ModuleName'])
            # just make sure different configurations do not cause errors in the output
            self.assertEqual(result.exit_code, 0)
            self.stop_server()

    def test_when_module_does_not_exist(self):
        server = FakeJoule()
        server.response = "module does not exist"
        server.http_code = 404
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['module', 'info', 'ModuleName'])
        self.assertIn("Error", result.output)
        self.stop_server()

    def test_when_server_returns_invalid_data(self):
        server = FakeJoule()
        server.response = "notjson"
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['module', 'info', 'ModuleName'])
        self.assertIn('Error', result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()

    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        server.response = "test error"
        server.http_code = 500
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['module', 'info', 'ModuleName'])
        self.assertIn('500', result.output)
        self.assertIn("test error", result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()



import aiodns
import aiohttp
from aiohttp import web
from sqlalchemy.orm import Session

from joule.models import Follower
from joule.api import TcpNode
from joule import app_keys
from joule.errors import ApiError

async def index(request: web.Request):
    db: Session = request.app[app_keys.db]
    followers = db.query(Follower)
    return web.json_response([f.to_json() for f in followers])

async def add(request: web.Request):
    """
    Called by a Joule node to allow *this* node to control it
    Parameters:
        key(str): the API key to use
        base_uri(str): (optional) specify the URL of the follower
        port(int): specify the port Joule is using on the follower

    """
    db: Session = request.app[app_keys.db]
    cafile = request.app[app_keys.cafile]
    if request.content_type != 'application/json':
        raise web.HTTPBadRequest(reason='content-type must be application/json')
    body = await request.json()

    try:
        key = body["key"]
        port = body["port"]
        scheme = body["scheme"]
        base_uri = body["base_uri"]
        # figure out the correct location of the follower
        if "name_is_host" in body:
            remote_host = body["name"]
        elif "X-Forwarded-For" in request.headers: # received from a reverse proxy
            remote_host = request.headers["X-Forwarded-For"]
        elif request.remote: # direct connection to built-in HTTP server
            remote_host = request.remote
        else:
            raise web.HTTPBadRequest(reason="could not determine remote host, specify with X-Forwarded-For if using a reverse proxy")
        location = scheme+"://" + remote_host + ":" + str(port) + base_uri
        
        node = TcpNode("new_follower", location, key, cafile)
        try:
            info = await node.info()
            follower_name = info.name
        except (aiohttp.ClientError, ApiError):
            raise web.HTTPBadRequest(reason=f"no route to node at {location}")
        finally:
            await node.close()
        follower = Follower(key=key, location=location,
                            name=follower_name)
        # update the follower if this is a repeat
        db.query(Follower).filter(Follower.location == follower.location).delete()
        db.add(follower)
        db.commit()
    except KeyError as e:
        raise web.HTTPBadRequest(reason="Invalid request, missing [%s]" % str(e))
    except ValueError as e:
        raise web.HTTPBadRequest(reason="Invalid request, bad argument format")
    return web.json_response(data={'name': request.app[app_keys.name]})


async def delete(request: web.Request):
    """
    Remove the specified node from the list of followers
    """
    db: Session = request.app[app_keys.db]
    try:
        name = request.query["name"]
        # if the name is not in the database return 404
        if db.query(Follower).filter(Follower.name == name).count() == 0:
            raise web.HTTPNotFound(reason="follower not found")
        db.query(Follower).filter(name == name).delete()
        db.commit()
    except KeyError:
        raise web.HTTPBadRequest(reason="specify name to remove")
    return web.Response(text="OK")



#!/usr/bin/env python3

from joule.utilities import time_now
import joule.client
import numpy as np
import asyncio
rows = 100
freq = 40  # Hz


class BrokenGen(joule.client.ReaderModule):

    async def run(self, parsed_args, output):
            data = 100 * np.sin(np.arange(0, 2 * np.pi, 2 * np.pi / rows))
            data.shape = (rows, 1)
            ts_inc = 1 / rows * (1 / freq) * 1e6  # microseconds
            data_ts = time_now()
            while not self.stop_requested:
                top_ts = data_ts + 100 * ts_inc
                ts = np.array(np.linspace(data_ts, top_ts, rows,
                                          endpoint=False), dtype=np.int64)
                ts.shape = (rows, 1)
                ts_data = np.hstack((ts, data))
                await output.write(ts_data)
                data_ts = top_ts
                await asyncio.sleep(1/freq)
                raise ValueError("Intentional Exception")


            
if __name__ == "__main__":
    r = BrokenGen()
    r.start()



import unittest
import traceback
import joule.errors
from joule import api, utilities, models
import numpy as np
import asyncio
import sys
from sqlalchemy import text


class TestDataMethods(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.node = api.get_node()

    async def asyncTearDown(self):
        await self.node.close()

    async def test_data_read(self):
        stream = await self.node.data_read("/live/base")
        num_intervals = 0
        # data should start at 0 and count up by 10's every interval
        while await stream.not_empty():
            data = await stream.read()
            # check the data
            expected = np.arange(0, len(data) * 10, 10)
            np.testing.assert_array_equal(data['data'], expected[:, None])
            if stream.end_of_interval:
                num_intervals += 1
            stream.consume(len(data))
        await stream.close()

    async def test_data_read_max_rows(self):
        stream = await self.node.data_read("/live/base", max_rows=100)
        total_rows = 0
        while await stream.not_empty():
            data = await stream.read()
            total_rows += len(data)
            stream.consume(len(data))
        self.assertLessEqual(total_rows, 105) # max rows is approximate, see joule/sql/row_count.sql

    async def test_data_subscribe(self):
        stream = await self.node.data_subscribe("/live/plus1")
        num_reads = 0
        TARGET_READS = 3
        MIN_DATA = 20
        # data should start at 0 and count up by 10's every interval
        while await stream.not_empty():
            data = await stream.read()
            await asyncio.sleep(0.5)
            num_reads += 1
            if num_reads >= TARGET_READS and len(data) > MIN_DATA:
                break
        stream.consume(len(data))
        np.testing.assert_array_almost_equal(
            np.ones((len(data) - 1)) * 10,
            np.diff(data['data'], axis=0).squeeze())

        await stream.close()

    async def test_data_write(self):
        time = utilities.time_now()
        print("deleting data...", end="")
        await self.node.data_delete("/archive/data1")
        print(f"OK {(utilities.time_now()-time)/1000:.2f}s")

        print("writing data...", end="")
        pipe = await self.node.data_write("/archive/data1")
        blk_size = 100
        nblks = 10
        _time = utilities.time_now()
        sys.stdout.flush()
        start_time = utilities.time_now()
        ts = np.linspace(start_time, start_time+ nblks*blk_size*100, nblks*blk_size, dtype=int)
        # make sure all timestamps are unique
        self.assertGreaterEqual(min(np.diff(ts)), 99)
        for i in range(nblks):
            # ts is a monotonic increasing sequence of timestamps counting by 100's
            ts_chunk = ts[i*blk_size:(i+1)*blk_size]
            self.assertEqual(len(ts_chunk), blk_size)
            data1 = np.random.random((blk_size, 1))
            data2 = np.random.random((blk_size, 1))
            res = np.hstack((ts_chunk[:, None], data1, data2))
            time = time + blk_size + 1
            await pipe.write(res)
        await pipe.close()
        print(f"OK {(utilities.time_now()-_time)/1000:.2f}s")

        print("reading data...", end="")
        time = utilities.time_now()
        # now read the data back
        pipe = await self.node.data_read("/archive/data1", start=0)
        nrows = 0
        while await pipe.not_empty():
            data = await pipe.read()
            nrows += len(data)
            pipe.consume(len(data))
        self.assertEqual(blk_size * nblks, nrows)
        print(f"OK {(utilities.time_now()-time)/1000:.2f}s")

    async def test_data_merges_successive_writes(self):
        await self.node.data_delete("/archive/data1")
        n = utilities.time_now()
        blk_size = 5
        ts = np.linspace(n+10e6, n+15e6, blk_size)
        data = np.random.random((blk_size, 2))
        pipe = await self.node.data_write("/archive/data1")
        #print([int(x) for x in ts])
        await pipe.write(np.hstack((ts[:, None], data)))
        await pipe.close()

        # now write some data that will *not* merge into the previous data interval
        ts = np.linspace(n+16e6, n+20e6, blk_size)
        pipe = await self.node.data_write("/archive/data1", merge_gap=1e3)
        #print([int(x) for x in ts])
        await pipe.write(np.hstack((ts[:, None], data)))
        await pipe.close()

        # now write some data that will merge into the previous data interval
        ts = np.linspace(n+21e6, n+25e6, blk_size)
        #print([int(x) for x in ts])
        pipe = await self.node.data_write("/archive/data1", merge_gap=2e6)
        await pipe.write(np.hstack((ts[:, None], data)))
        await pipe.close()

        # expect 2 intervals
        intervals = await self.node.data_intervals("/archive/data1")
        #print(intervals)
        self.assertEqual(len(intervals), 2)

        # now read the data back
        pipe = await self.node.data_read("/archive/data1", start=0)
        nrows = 0
        first_block = True
        #print("===== READING DATA ======")
        while await pipe.not_empty():
            data = await pipe.read()
            #print([int(x) for x in data['timestamp']])
            nrows += len(data)
            #if pipe.end_of_interval:
            #    print("--- interval break ---")
            if pipe.end_of_interval and first_block:
                # 1st block has one chunk of data
                self.assertEqual(len(data), blk_size)
                first_block = False
                nrows=0
            pipe.consume(len(data))
        # 2nd block has two chunks of data
        self.assertFalse(first_block) # make sure we already read the first block
        self.assertEqual(nrows, blk_size*2)

    async def test_data_write_invalid_values(self):
        # write duplicate data to a pipe
        await self.node.data_delete("/archive/data1")
        pipe = await self.node.data_write("/archive/data1")
        blk_size = 100
        time = utilities.time_now()
        ts = np.linspace(time, time + blk_size, blk_size)
        self.assertEqual(min(np.diff(ts)), 1)
        data1 = np.random.random((blk_size, 1))
        data2 = np.random.random((blk_size, 1))
        res = np.hstack((ts[:, None], data1, data2))
        await pipe.write(res[75:, :])
        try:
            await pipe.write(res[:50, :])
            await pipe.close()

            self.fail("should have raised an exception")
        except joule.errors.ApiError as e:
            assert "monotonic" in traceback.format_exc()
            pass
        # write duplicate data across two pipes
        await self.node.data_delete("/archive/data1")
        pipe = await self.node.data_write("/archive/data1")
        await pipe.write(res[25:, :])
        await pipe.close()
        pipe = await self.node.data_write("/archive/data1")
        try:
            await pipe.write(res[:50, :])
            await pipe.close()
            self.fail("should have raised an exception")
        except joule.errors.ApiError as e:
            assert "already exists" in traceback.format_exc()
        # write data with NaN's
        await self.node.data_delete("/archive/data1")
        pipe = await self.node.data_write("/archive/data1")
        res[0, 2] = np.nan
        try:
            await pipe.write(res)
            await pipe.close()
            self.fail("should have raised an exception")
        except joule.errors.ApiError as e:
            assert "NaN" in traceback.format_exc()

    async def test_data_delete(self):
        # copy data to a new stream
        src = await self.node.data_stream_get("/live/base")
        # wait for source stream to have data in it
        while True:
            src_info = await self.node.data_stream_info(src)
            if src_info.rows > 10:
                break
            await asyncio.sleep(1)  # 10Hz rate so this should just happen once
        dest = await self.node.data_stream_create(src, "/tmp")
        data_in = await self.node.data_read(src)
        data_out = await self.node.data_write(dest)
        first_ts = None
        last_ts = None
        while await data_in.not_empty():
            data = await data_in.read()
            if first_ts is None:
                first_ts = data['timestamp'][0]
            last_ts = data['timestamp'][-1]
            data_in.consume(len(data))
            await data_out.write(data)

        await data_out.close()
        dest_info = await self.node.data_stream_info(dest)
        self.assertGreater(dest_info.rows, 2)
        # remove all but the first and last samples
        await self.node.data_delete(dest, start=first_ts + 1, end=last_ts - 1)
        dest_info = await self.node.data_stream_info(dest)
        self.assertEqual(dest_info.rows, 2)
        await self.node.data_stream_delete(dest)



from aiohttp import web
from joule.models import DataStore
import joule
import pkg_resources
from joule import app_keys
async def index(request: web.Request):
    return web.Response(text="Joule server")


async def db_connection(request: web.Request):
    return web.json_response(data=
                             request.app[app_keys.module_connection_info].to_json())


async def dbinfo(request: web.Request):
    data_store: DataStore = request.app[app_keys.data_store]
    data = await data_store.dbinfo()
    return web.json_response(data=data.to_json())


async def version_json(request: web.Request):
    return web.json_response(data={'version': pkg_resources.get_distribution('joule').version,
                                   'name': request.app[app_keys.name],
                                   'uuid': str(request.app[app_keys.uuid])})


async def version(request: web.Request):
    return web.Response(text=pkg_resources.get_distribution('joule').version)



import click

from joule.cli.lazy_group import LazyGroup

@click.group(name="stream",
             cls=LazyGroup,
             lazy_subcommands={
                 "info": "joule.cli.stream.info.cli_info",
                 "move": "joule.cli.stream.move.cli_move",
                 "delete": "joule.cli.stream.destroy.cli_delete",
                 "annotations": "joule.cli.stream.annotation.cli_annotations",
                 "rename": "joule.cli.stream.rename.cli_rename"
             })
def streams():
    """Manage data streams."""
    pass  






#!/usr/bin/env -S python3 -u

"""
Run through each scenario in scenarios
"""
import os
import sys
import subprocess
import shlex
import shutil
import signal
import asyncio
from helpers import wait_for_joule_host
import joule
import time
from joule import api

SOURCE_DIR = "/joule"
SCENARIO_DIR = "/joule/tests/e2e/scenarios"
MODULE_SCRIPT_DIR = "/joule/tests/e2e/module_scripts"
JOULE_CONF_DIR = "/etc/joule"
SECURITY_DIR = "/joule/tests/e2e/pki"

JOULED_CMD ="coverage run --rcfile=/joule/.coveragerc -m joule.daemon".split(" ")
JOULE_CMD = "coverage run --rcfile=/joule/.coveragerc -m joule.cli".split(" ")

FORCE_DUMP = True


async def wait_for_follower():
    node1 = api.get_node("node1.joule")
    try_count=0
    while try_count < 25:
        try_count += 1
        try:
            followers = await node1.follower_list()
            if len(followers) == 1:
                return 0 # success
            else:
                await asyncio.sleep(1)
        except joule.errors.ApiError:
            # wait until node is online
            await asyncio.sleep(1)
        finally:
            await node1.close()
    raise SystemExit("follower did not connect")


def main():
    tests = []
    for entry in os.scandir(SCENARIO_DIR):
        if not (entry.name.startswith('.') and entry.is_dir()):
            if entry.name in ["__pycache__", "__init__.py"]:
                continue
            tests.append((entry.name, entry.path))

    # override to run only particular test(s)
    # tests = [("multinodes", "/joule/tests/e2e/scenarios/multiple_nodes")]
    # tests = [("standalone modules", "/joule/tests/e2e/scenarios/standalone_modules")]
    # tests = [("API", "/joule/tests/e2e/scenarios/api_tests")]
    # tests = [("Basic Operation", "/joule/tests/e2e/scenarios/basic_operation")]
    # tests = [("Data Management", "/joule/tests/e2e/scenarios/data_management")]
    # tests = [("Quick Start Pipeline", "/joule/tests/e2e/scenarios/quick_start_pipeline")]
    # tests = [] # don't run any tests
    # make module scripts available
    os.symlink(MODULE_SCRIPT_DIR, "/module_scripts")

    # get API permissions
    os.environ["LOGNAME"] = "e2e"
    os.environ["JOULE_USER_CONFIG_DIR"] = "/tmp/joule_user"

    # wait for the follower to connect
    ### NOTE: To debug problems with jouled initializing, comment out the stdout and stderr redirects below
    jouled = subprocess.Popen(JOULED_CMD,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    universal_newlines=True)
    time.sleep(2) # wait for jouled to start

    ## The main node tests out the NGinx proxy functionality (follower uses built-in HTTP server)
    subprocess.run(JOULE_CMD+"admin authorize --url=http://localhost/joule".split(" "))
    wait_for_joule_host('node1.joule')
    asyncio.run(wait_for_follower())
    jouled.send_signal(signal.SIGINT)

    ## Add a user through the management file interface
    ## this is verified by the API users test
    with open("/etc/joule/users.conf", "w") as f:
        f.write("auto_user, 044b1a5153e5f736cd787870cc9bcf2a\n")

    for (test_name, test_path) in tests:
        print("---------[%s]---------" % test_name)
        # remove the contents of /etc/joule/module_configs and replace with this test's configs
        shutil.rmtree("/etc/joule/module_configs")
        shutil.rmtree("/etc/joule/stream_configs")
        shutil.copytree(os.path.join(test_path, "module_configs"), "/etc/joule/module_configs")
        shutil.copytree(os.path.join(test_path, "stream_configs"), "/etc/joule/stream_configs")

        # clear the existing database (keeping master/follower tables)
        with open(os.devnull, 'w') as devnull:
            subprocess.run(JOULE_CMD+"admin erase --yes".split(" "), stdout=devnull)

        # start jouled
        jouled = subprocess.Popen(JOULED_CMD,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT,
                                  universal_newlines=True)
        wait_for_joule_host('node1.joule')
        time.sleep(15)  # wait for jouled to boot and collect data

        sys.stdout.flush()
        test = subprocess.run(("coverage run --rcfile=/joule/.coveragerc "+os.path.join(test_path, "test.py")).split(" "))
        jouled.send_signal(signal.SIGINT)
        stdout, _ = jouled.communicate()
        if test.returncode != 0 or FORCE_DUMP:
            print("----dump from jouled----")
            for line in stdout.rstrip().split('\n'):
                print("> %s" % line)
            if test.returncode != 0:
                return test.returncode
    return 0  # success


if __name__ == "__main__":
    exit(main())



import asyncio
from typing import List, Optional, Callable, Coroutine, Dict
import numpy as np
from io import BytesIO
import asyncpg
import asyncpg.exceptions
import datetime
import shutil
import asyncio
from joule.utilities import interval_tools, timestamp_to_datetime
from joule.models.data_store.timescale_inserter import Inserter, Decimator
from joule.models import DataStream, pipes
from joule.models.data_store import psql_helpers
from joule.models.data_store.data_store import DataStore, StreamInfo, DbInfo
from joule.utilities import datetime_to_timestamp as dt2ts
Loop = asyncio.AbstractEventLoop


class TimescaleStore(DataStore):

    def __init__(self, dsn: str, insert_period: float,
                 cleanup_period: float):
        self.dsn = dsn
        self.decimation_factor = 4
        self.insert_period = insert_period
        self.cleanup_period = cleanup_period
        self.pool = None
        # TimescaleDB supports decimation management
        self.supports_decimation_management = True 
        # tunable constants
        self.extract_block_size = 50000

    async def initialize(self, streams: List[DataStream]) -> None:
        self.pool = await asyncpg.create_pool(self.dsn, command_timeout=60)
        return self.pool  # so the event store can reuse it

    async def close(self):
        try:
            await asyncio.wait_for(self.pool.close(), timeout=5.0)
        except asyncio.TimeoutError:
            print("WARNING: could not close timescale pool, terminating")

    async def spawn_inserter(self, stream: 'DataStream', pipe: pipes.Pipe, insert_period=None,
                             merge_gap=0) -> asyncio.Task:
        if insert_period is None:
            insert_period = self.insert_period

        inserter = Inserter(self.pool, stream,
                            insert_period=insert_period, 
                            cleanup_period=self.cleanup_period,
                            merge_gap=merge_gap)
        t = asyncio.create_task(inserter.run(pipe))
        t.set_name("Timescale Inserter for [%s]" % stream.name)
        return t

    async def insert(self, stream: 'DataStream',
                     data: np.ndarray, start: int, end: int):
        print("WARNING: insert called on timescale store this is a NOP, use spawn_inserter instead")

    async def drop_decimations(self, stream: 'DataStream'):
        async with self.pool.acquire() as conn:
            await psql_helpers.drop_decimation_tables(conn, stream)

    async def decimate(self, stream: 'DataStream'):
        await self.drop_decimations(stream)
        async with self.pool.acquire() as conn:
            # do not change the chunk interval sizes
            decimator = Decimator(stream=stream, from_level=1, factor=4, chunk_interval=0)
            interval_token = pipes.interval_token(stream.layout)
            async def redecimator(data, _, __):
                if len(data)==0:
                    print("WARNING: decimator received a 0 length data array")
                    return
                # if the data is just an interval token, close the interval
                if len(data)==1 and data[0] == interval_token:
                    decimator.close_interval()
                    return
                # if the end of the data is an interval token, remove it
                if data[-1] == interval_token:
                    await decimator.process(conn, data[:-1])
                    decimator.close_interval()
                else: # normal data just process it
                    await decimator.process(conn, data)

            # start a data extraction task with the callback running
            # the decimation, process the whole stream
            start, end  = await psql_helpers.convert_time_bounds(conn, stream, None, None)
            await _extract_data(conn, stream, redecimator,
                                decimation_level=1, start=start, end=end,
                                block_size=50000)

    async def consolidate(self, stream: 'DataStream', start: int,
                          end: int, max_gap: int) -> int:
        # remove interval gaps less than or equal to max_gap duration (in us)
        intervals = await self.intervals(stream, start, end)
        if len(intervals) == 0:
            return 0  # no data, nothing to do
        duration = [intervals[0][0], intervals[-1][1]]

        gaps = interval_tools.interval_difference([duration], intervals)

        if len(gaps) == 0:
            return 0  # no interval breaks, nothing to do
        small_gaps = [gap for gap in gaps if (gap[1] - gap[0]) <= max_gap]
        if len(small_gaps) == 0:
            return 0  # no small gaps, nothing to do
        boundaries = [gap[0] for gap in small_gaps]
        str_datetimes = ["'%s'" % str(timestamp_to_datetime(ts)) for ts in boundaries]
        query = "DELETE FROM data.stream%d_intervals WHERE time IN (%s)" % (stream.id, ",".join(str_datetimes))
        async with self.pool.acquire() as conn:
            await conn.execute(query)
        return len(small_gaps)

    async def intervals(self, stream: 'DataStream', start: Optional[int], end: Optional[int]):
        # find the intervals
        # for each interval find the nearest data
        intervals = []
        async with self.pool.acquire() as conn:
            boundaries = await psql_helpers.get_boundaries(conn, stream, start, end)
            cur_start = None
            for i in range(len(boundaries)):
                boundary = boundaries[i]
                if cur_start is not None:
                    query = "SELECT time FROM data.stream%d WHERE time < '%s'" % (stream.id, boundary)
                    query += " ORDER BY time DESC LIMIT 1"
                    prev_ts = await conn.fetchval(query)
                    utc_start_ts = round(cur_start.replace(tzinfo=datetime.timezone.utc).timestamp() * 1e6)
                    utc_end_ts = round(prev_ts.replace(tzinfo=datetime.timezone.utc).timestamp() * 1e6)
                    # intervals are [..) with extra us on the end
                    intervals.append([utc_start_ts, utc_end_ts + 1])
                    cur_start = None
                query = "SELECT time FROM data.stream%d WHERE time >= '%s'" % (stream.id, boundary)
                if i < (len(boundaries) - 1):
                    query += " AND time < '%s' " % (boundaries[i + 1])
                query += " ORDER BY time ASC LIMIT 1"
                next_ts = await conn.fetchval(query)
                if next_ts is None:
                    # unecessary interval marker, remove it?
                    continue
                if cur_start is None:
                    cur_start = next_ts
        return intervals

    async def extract(self, stream: 'DataStream', start: Optional[int], end: Optional[int],
                      callback: Callable[[np.ndarray, str, int], Coroutine],
                      max_rows: int = None, decimation_level=None):
        conn = await self.pool.acquire()
        # limit time bounds to range of base stream
        (start, end) = await psql_helpers.limit_time_bounds(conn, stream, start, end)
        # figure out appropriate decimation level
        if decimation_level is None:
            if max_rows is None:
                decimation_level = 1
            else:
                # find out how much data this represents
                count = await psql_helpers.get_row_count(conn, stream, start, end)
                if count > 0:
                    desired_decimation = np.ceil(count / max_rows)

                    decimation_level = int(4 ** np.ceil(np.log(desired_decimation) /
                                                        np.log(self.decimation_factor)))
                    # print("count=%d, max_rows=%d,desired_decim=%d,decim_level=%d" % (
                    #    count, max_rows, desired_decimation, decimation_level))
                else:
                    # create an empty array with the right data type
                    data = np.array([], dtype=pipes.compute_dtype(stream.layout))
                    await callback(data, stream.layout, 1)
                    await self.pool.release(conn)
                    return
        try:
            await _extract_data(conn, stream, callback, decimation_level, start, end,
                                block_size=self.extract_block_size)
        finally:
            await self.pool.release(conn)

    async def remove(self, stream: 'DataStream',
                     start: Optional[int], end: Optional[int],
                     exact: bool = True):
        where_clause = psql_helpers.query_time_bounds(start, end)
        if len(where_clause)>0:
            where_clause = " WHERE " + where_clause
        async with self.pool.acquire() as conn:
            tables = await psql_helpers.get_table_names(conn, stream)
            for table in tables:
                # TODO: use drop chunks with newer and older clauses when timescale is updated
                if start is None and end is None:
                    query = "TRUNCATE %s" % table
                elif start is None and "intervals" not in table and not exact:
                    # use the much faster drop chunks utility and accept the approximate result
                    bounds = await psql_helpers.convert_time_bounds(conn, stream, start, end)
                    if bounds is None:
                        return  # no data to remove
                    query = "SELECT drop_chunks('%s', older_than=>'%s'::timestamp)" % (table, bounds[1])
                else:
                    query = f'DELETE FROM {table} {where_clause}'
                try:
                    await conn.execute(query)
                except asyncpg.UndefinedTableError:
                    return  # no data to remove
                except asyncpg.exceptions.RaiseError as err:
                    print("psql: ", err)
                    return
            # create an interval boundary to mark the missing data
            if start is not None:
                await psql_helpers.close_interval(conn, stream, start)

    async def destroy(self, stream: 'DataStream'):
        async with self.pool.acquire() as conn:
            tables = await psql_helpers.get_table_names(conn, stream)
            for table in tables:
                try:
                    await conn.execute('DROP TABLE %s ' % table)
                except asyncpg.UndefinedTableError:
                    pass

    async def destroy_all(self):
        async with self.pool.acquire() as conn:
            tables = await psql_helpers.get_all_table_names(conn)
            for table in tables:
                try:
                    await conn.execute('DROP TABLE %s ' % table)
                except asyncpg.UndefinedTableError:
                    pass

    async def info(self, streams: List['DataStream']) -> Dict[int, StreamInfo]:
        results = {}
        async with self.pool.acquire() as conn:
            for my_stream in streams:
                try:
                    record = await conn.fetchrow("select * from stream_info(%d);" % my_stream.id)
                except asyncpg.UndefinedTableError:
                    # this stream has no data tables, just make up an empty record
                    record = {'min_ts': None, 'max_ts': None, 'rows': 0, 'size': 0}
                start = record['min_ts']
                end = record['max_ts']

                if start is not None:
                    start = round(start.replace(tzinfo=datetime.timezone.utc)
                                .timestamp() * 1e6)
                if end is not None:
                    end = round(end.replace(tzinfo=datetime.timezone.utc)
                              .timestamp() * 1e6)
                if start is not None and end is not None:
                    total_time = end - start
                else:
                    total_time = 0
                results[my_stream.id] = StreamInfo(start, end,
                                                   record['rows'],
                                                   total_time, record['size'])
        return results

    async def dbinfo(self) -> DbInfo:

        async with self.pool.acquire() as conn:
            dbsize = await conn.fetchval("select pg_database_size(current_database())")  # in bytes
            path = await conn.fetchval("show data_directory")
            try:
                usage = shutil.disk_usage(path)  # usage in bytes
                return DbInfo(path=path,
                              size=dbsize,
                              other=max(usage.used - dbsize, 0),
                              reserved=max(usage.total - usage.used - usage.free, 0),
                              free=usage.free)
            except FileNotFoundError:
                # this occurs if the database is remote
                return DbInfo(path="--remote-database--",
                              size=-1,
                              other=-1,
                              reserved=-1,
                              free=-1)


async def _extract_data(conn: asyncpg.Connection, stream: DataStream, callback,
                        decimation_level: int = 1, start: Optional[int] = None, end: Optional[int] = None,
                        block_size=50000):
    if decimation_level > 1:
        layout = stream.decimated_layout
    else:
        layout = stream.layout

    table_name = "data.stream%d" % stream.id
    if decimation_level > 1:
        table_name += "_%d" % decimation_level
    # extract by interval
    query = "SELECT time FROM data.stream%d_intervals " % stream.id
    where_clause = psql_helpers.query_time_bounds(start, end)
    if len(where_clause)>0:
        query += " WHERE " + where_clause
    query += " ORDER BY time ASC"
    try:
        data = await conn.fetch(query)
        # convert to timestamps
        boundary_records = [dt2ts(row['time']) for row in data]
    except asyncpg.UndefinedTableError:
        # no data tables
        data = np.array([], dtype=pipes.compute_dtype(layout))
        await callback(data, layout, decimation_level)
        return

    boundary_records.append(end)
    for i in range(len(boundary_records)):
        end = boundary_records[i]
        # extract the interval data
        if start==end: # do not run an empty query
            continue
        done = False
        while not done:
            query = "SELECT * FROM %s " % table_name
            where_clause = psql_helpers.query_time_bounds(start, end)
            if len(where_clause)>0:
                query+= " WHERE " + where_clause
            query += " ORDER BY time ASC LIMIT %d" % block_size
            psql_bytes = BytesIO()
            try:
                await conn.copy_from_query(query, format='binary', output=psql_bytes)
            except asyncpg.UndefinedTableError:
                # interval table exists but not the data table
                data = np.array([], dtype=pipes.compute_dtype(layout))
                await callback(data, layout, decimation_level)
                return
            psql_bytes.seek(0)
            dtype = pipes.compute_dtype(layout)
            np_data = psql_helpers.bytes_to_data(psql_bytes, dtype)

            if len(np_data) < block_size:
                # this is the last block in the interval, append an interval token
                # unless this is the end of the requested data
                if i < len(boundary_records) - 1:
                    # add an interval token to the end of np_data
                    np_data = np.append(np_data, pipes.interval_token(layout))
            if len(np_data)==0:
                print("WARNING: empty read, no data found in interval")
                break
                
            await callback(np_data, layout, decimation_level)

            if len(np_data) < block_size:
                break
            start = np_data['timestamp'][-1] + 1
        # do not put an interval token at the end of the data
        #if i < len(boundary_records) - 1:
        #    await callback(pipes.interval_token(layout), layout, decimation_level)
        start = end



import unittest
from sqlalchemy import create_engine
from sqlalchemy.orm import Session

import logging
import tempfile
import os

from tests.helpers import DbTestCase
from joule.services import load_modules

logger = logging.getLogger('joule')

class TestLoadModulesErrors(DbTestCase):

    def test_module_must_have_exec_cmd(self):
        conf_str = """
                            [Main]
                            name=bad module
                            [Inputs]
                            input=/path/to/input:float32[y]
                            [Outputs]
                            output=/path/to/output:float32[x]
                            """
        with self.assertLogs(level="ERROR") as logs:
            attempt_load(conf_str, self.db)
        all_logs = ' '.join(logs.output).lower()
        self.assertIn('exec_cmd', all_logs)

    def test_streams_must_be_configured(self):
        conf_str = """
                    [Main]
                    name=bad module
                    exec_cmd=runit.sh
                    [Outputs]
                    output=/path/not/configured
                    # no inputs
                    """
        with self.assertLogs(level="ERROR") as logs:
            attempt_load(conf_str, self.db)
        all_logs = ' '.join(logs.output).lower()
        self.assertIn('configured', all_logs)


def attempt_load(conf_str, db: Session):
    with tempfile.TemporaryDirectory() as conf_dir:
        with open(os.path.join(conf_dir, "invalid.conf"), 'w') as f:
            f.write(conf_str)
        load_modules.run(conf_dir, db)



from click.testing import CliRunner

import warnings
import numpy as np
import logging
import tempfile
import h5py
import json
from datetime import datetime, timezone
import unittest
from icecream import ic
from ..fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main
from joule.models import DataStream, Element, StreamInfo
from joule.api.data_stream import Element as ApiElement
from tests import helpers

warnings.simplefilter('always')
log = logging.getLogger('aiohttp.access')
log.setLevel(logging.WARNING)


class TestDataIngest(FakeJouleTestCase):

    def test_ingests_data_to_empty_existing_stream(self):
        server = FakeJoule()
        # create the source stream
        src = DataStream(id=0, name="existing", keep_us=100,
                         datatype=DataStream.DATATYPE.FLOAT32, updated_at=datetime.now(timezone.utc))
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]
        src_data = helpers.create_data(src.layout, length=22000)
        src_info = StreamInfo(0, 0, 0, 0)
        server.add_stream('/test/existing', src, src_info, None)
        self.start_server(server)
        runner = CliRunner()
        with tempfile.NamedTemporaryFile() as data_file:
            write_hd5_data(data_file, src_data)
            result = runner.invoke(main, ['data', 'ingest', data_file.name, '--stream', '/test/existing'])
            _print_result_on_error(result)
            self.assertEqual(result.exit_code, 0)
        db_obj = self.msgs.get()
        np.testing.assert_array_equal(src_data, db_obj.data)
        # uses the stream parameter instead of the hd5 attrs
        self.assertEqual(db_obj.stream.name, 'existing')
        self.stop_server()

    def test_datatype_mismatch(self):
        #  the datatype of the file and target stream must match
        server = FakeJoule()
        # create the source stream
        src = DataStream(id=0, name="dest", keep_us=100,
                         datatype=DataStream.DATATYPE.INT16, updated_at=datetime.now(timezone.utc))
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]
        # source has 100 rows of data between [0, 100]
        file_data = helpers.create_data('int32_3')
        src_info = StreamInfo(0, 0, 0, 0)
        server.add_stream('/test/dest', src, src_info, None)
        self.start_server(server)
        runner = CliRunner()
        with tempfile.NamedTemporaryFile() as data_file:
            write_hd5_data(data_file, file_data)
            result = runner.invoke(main, ['data', 'ingest', data_file.name])
            self.assertIn("datatype", result.output)
            self.assertNotEqual(result.exit_code, 0)

        self.stop_server()

    def test_confirms_data_removal(self):
        # if there is existing data in the target, confirm removal
        server = FakeJoule()
        # create the source stream
        src = DataStream(id=0, name="dest", keep_us=100,
                         datatype=DataStream.DATATYPE.FLOAT32, updated_at=datetime.now(timezone.utc))
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]
        # source has 100 rows of data between [0, 100]
        src_data = helpers.create_data(src.layout, length=1000)
        # File:    |----|
        # DataStream:     |-----|
        src_info = StreamInfo(int(src_data['timestamp'][500]),
                              int(src_data['timestamp'][-1]),
                              500)
        server.add_stream('/test/dest', src, src_info, src_data[500:])
        self.start_server(server)
        runner = CliRunner()
        with tempfile.NamedTemporaryFile() as data_file:
            write_hd5_data(data_file, src_data[:750])
            result = runner.invoke(main, ['data', 'ingest', data_file.name],
                                   input='N')
            _print_result_on_error(result)
            self.assertEqual(result.exit_code, 0)
            self.assertIn('Cancelled', result.output)

        self.stop_server()

    def test_dataset_mismatch(self):
        # the length of [data] and [timestamp]  must match
        runner = CliRunner()
        with tempfile.NamedTemporaryFile() as data_file:
            with h5py.File(data_file.name, 'w') as hdf_root:
                hdf_root.create_dataset('data', (10, 10),
                                        dtype=np.dtype('float32'),
                                        compression='gzip')
                hdf_root.create_dataset('timestamp', (11, 1),
                                        dtype=np.dtype('float32'),
                                        compression='gzip')
            result = runner.invoke(main, ['data', 'ingest', data_file.name])
            self.assertIn("Error", result.output)
            self.assertNotEqual(result.exit_code, 0)

    def test_stream_element_mismatch(self):
        #  the datatype of the file and target stream must match
        server = FakeJoule()
        # create the source stream
        src = DataStream(id=0, name="dest", keep_us=100,
                         datatype=DataStream.DATATYPE.INT16, updated_at=datetime.now(timezone.utc))
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(4)]
        # source has 100 rows of data between [0, 100]
        file_data = helpers.create_data('int16_3')
        src_info = StreamInfo(0, 0, 0, 0)
        server.add_stream('/test/dest', src, src_info, None)
        self.start_server(server)
        runner = CliRunner()
        with tempfile.NamedTemporaryFile() as data_file:
            write_hd5_data(data_file, file_data)
            result = runner.invoke(main, ['data', 'ingest', data_file.name])
            self.assertIn("elements", result.output)
            self.assertNotEqual(result.exit_code, 0)

        self.stop_server()

    def test_creates_stream_from_attributes(self):
        # creates the target stream based on hdf attrs
        server = FakeJoule()
        src_data = helpers.create_data('float32_3')
        self.start_server(server)
        runner = CliRunner()

        with tempfile.NamedTemporaryFile() as data_file:
            write_hd5_data(data_file, src_data[:750])
            result = runner.invoke(main, ['data', 'ingest', data_file.name],
                                   input='N')
            _print_result_on_error(result)
            self.assertEqual(result.exit_code, 0)
            # prints a message about the new stream
            self.assertIn('/test/dest', result.output)
        db_entry = self.msgs.get()
        # make sure all the elements are present
        elem_names = [e.name for e in db_entry.stream.elements]
        expected_names = ['Test %d' % i for i in range(3)]
        self.assertEqual(elem_names, expected_names)
        self.assertEqual(db_entry.stream.name, 'dest')
        self.stop_server()

    def test_invalid_data_structure(self):
        runner = CliRunner()
        with tempfile.NamedTemporaryFile() as data_file:
            with h5py.File(data_file.name, 'w') as hdf_root:
                hdf_root.create_dataset('data', (1, 10),
                                        dtype=np.dtype('float32'),
                                        compression='gzip')

            result = runner.invoke(main, ['data', 'ingest', data_file.name])
            self.assertIn("Error", result.output)
            self.assertNotEqual(result.exit_code, 0)

    def test_invalid_filetype(self):
        runner = CliRunner()
        with tempfile.NamedTemporaryFile() as bad_file:
            with open(bad_file.name, 'w') as f:
                f.write('invalid filetype')
            result = runner.invoke(main, ['data', 'ingest', bad_file.name])
            self.assertIn("Error", result.output)
            self.assertNotEqual(result.exit_code, 0)


def write_hd5_data(path, data):
    with h5py.File(path, 'w') as hdf_root:
        hdf_data = hdf_root.create_dataset('data', data['data'].shape,
                                           dtype=data.dtype[1].base,
                                           compression='gzip')
        hdf_timestamps = hdf_root.create_dataset('timestamp', [len(data), 1],
                                                 dtype='i8',
                                                 compression='gzip')
        hdf_data[...] = data['data']
        hdf_timestamps[...] = data['timestamp'][:, None]
        hdf_root.attrs['path'] = '/test/dest'
        elements = [ApiElement(name='Test %d' % i,
                               display_type='discrete') for i in range(data['data'].shape[1])]
        element_json = json.dumps([e.to_json() for e in elements])
        hdf_root.attrs['element_json'] = element_json


def _print_result_on_error(result):
    if result.exit_code != 0:
        print("output: ", result.output)
        print("exception: ", result.exception)



import os
import argparse
import asyncio
import json
import io
import numpy as np
from contextlib import redirect_stdout
from aiohttp import web
import aiohttp
from aiohttp.test_utils import unused_port
import threading
import time
import requests

from joule.client import ReaderModule
from joule.models import DataStream, Element, pipes
from tests import helpers
import warnings

warnings.simplefilter('always')


class SimpleReader(ReaderModule):
    async def run(self, parsed_args, output: pipes.Pipe):
        await output.write(parsed_args.mock_data)
        await output.close()


class InterfaceReader(ReaderModule):
    async def run(self, parsed_args, output: pipes.Pipe):
        while not self.stop_requested:
            await asyncio.sleep(0.01)

    def routes(self):
        return [web.get('/', self.index)]

    async def index(self, _):
        self.stop_requested = True
        return web.Response(text="Hello World")


class TestReaderModule(helpers.AsyncTestCase):

    def setUp(self):
        super().setUp()
        # module output is a float32_3 stream
        self.stream = DataStream(name="output", datatype=DataStream.DATATYPE.FLOAT32,
                                 elements=[Element(name="e%d" % j, index=j,
                                               display_type=Element.DISPLAYTYPE.CONTINUOUS) for j in range(3)])

    def test_writes_to_pipes(self):
        module = SimpleReader()
        (r, w) = os.pipe()

        rf = pipes.reader_factory(r)
        pipe = pipes.InputPipe(name="output", stream=self.stream, reader_factory=rf)
        pipe_arg = json.dumps(json.dumps({"outputs": {'output': {'fd': w, 'id': None, 'layout': self.stream.layout}},
                                          "inputs": {}}))
        data = helpers.create_data(self.stream.layout)
        args = argparse.Namespace(pipes=pipe_arg, socket="unset",
                                  url='http://localhost:8080',
                                  node="", api_socket="",
                                  mock_data=data)
        # run the reader module

        module.start(args)
        # check the output
        received_data = asyncio.run(pipe.read())
        np.testing.assert_array_equal(data, received_data)
        asyncio.run(pipe.close())

    def test_writes_to_stdout(self):
        module = SimpleReader()
        data = helpers.create_data(self.stream.layout, length=10)
        args = argparse.Namespace(pipes="unset",
                                  module_config="unset",
                                  url='http://localhost:8080',
                                  node="", api_socket="",
                                  socket="unset", mock_data=data)
        # run the reader module
        f = io.StringIO()
        with redirect_stdout(f):
            module.start(args)
        lines = f.getvalue().split('\n')
        for i in range(len(data)):
            output = lines[i].split(' ')
            ts = int(output[0])
            rx_data = [float(x) for x in output[1:]]
            self.assertEqual(ts, data['timestamp'][i])
            np.testing.assert_array_almost_equal(rx_data, data['data'][i])

    def test_reader_requires_single_output(self):
        module = SimpleReader()
        data = helpers.create_data(self.stream.layout, length=10)
        pipe_arg = json.dumps(json.dumps({"outputs": {'first': {'fd': 0, 'id': None, 'layout': self.stream.layout},
                                                      'second': {'fd': 1, 'id': None, 'layout': self.stream.layout}},
                                          "inputs": {}}))
        args = argparse.Namespace(pipes=pipe_arg,
                                  module_config="unset",
                                  socket="unset",
                                  url='http://localhost:8080',
                                  node="", api_socket="",
                                  mock_data=data)
        # run the reader module
        with self.assertLogs(level="ERROR") as logs:
            module.start(args)
        all_logs = ' '.join(logs.output).lower()
        self.assertTrue('output' in all_logs)

    def test_runs_webserver(self):
        module = InterfaceReader()
        data = helpers.create_data(self.stream.layout, length=10)
        port = unused_port()
        args = argparse.Namespace(pipes="unset", module_config="unset",
                                  socket="unset", port=port, host="127.0.0.1",
                                  url='http://localhost:8080',
                                  node="", api_socket="",
                                  mock_data=data)

        def get_page():
            time.sleep(0.5)
            resp = requests.get('http://localhost:%d' % port)
            self.assertEqual(resp.content.decode('utf8'), 'Hello World')

        getter = threading.Thread(target=get_page)
        getter.start()
        f = io.StringIO()
        with redirect_stdout(f):
            module.start(args)
        getter.join()

    def test_opens_socket(self):
        socket_path = '/tmp/interface.test'
        if os.path.exists(socket_path):
            os.unlink(socket_path)

        module = InterfaceReader()
        data = helpers.create_data(self.stream.layout, length=10)
        args = argparse.Namespace(pipes="unset", module_config="unset",
                                  url='http://localhost:8080',
                                  socket=socket_path,
                                  node="", api_socket="",
                                  mock_data=data)

        def get_page():
            time.sleep(0.5)
            resp = asyncio.run(get_page_task())
            self.assertEqual(resp, 'Hello World')
            

        async def get_page_task():
            conn = aiohttp.UnixConnector(socket_path)
            async with aiohttp.ClientSession(connector=conn,
                                             auto_decompress=False) as session:
                # proxy the request to the module
                async with session.get('http://localhost/') as resp:
                    return await resp.text()

        getter = threading.Thread(target=get_page)
        getter.start()
        f = io.StringIO()
        with redirect_stdout(f):
            module.start(args)
        getter.join()
        if os.path.exists(socket_path):
            os.unlink(socket_path)



from click.testing import CliRunner
import warnings
import numpy as np
import logging
from datetime import datetime, timezone
import asyncio
import unittest

from ..fake_joule import FakeJoule, FakeJouleTestCase, MockDbEntry
from joule.cli import main
from joule.models import DataStream, Element, StreamInfo, pipes
from tests import helpers
import time

warnings.simplefilter('always')
log = logging.getLogger('aiohttp.access')
log.setLevel(logging.WARNING)


class TestDataCopy(FakeJouleTestCase):

    def test_copies_all_data(self):
        server = FakeJoule()
        # create the source and destination streams
        src_data = create_source_data(server, break_into_intervals=False)  # helpers.create_data(src.layout)
        # dest is empty
        dest = DataStream(id=1, name="dest", keep_us=100,
                          datatype=DataStream.DATATYPE.FLOAT32, updated_at=datetime.now(timezone.utc))
        dest.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in
                         range(3)]
        server.add_stream('/test/destination', dest, StreamInfo(None, None, 0), None)

        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'copy', '/test/source', '/test/destination'])
        _print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        # wait for the copy to complete
        while self.msgs.empty():
            time.sleep(0.1)
        max_msgs = 40
        num_msgs = 0
        mock_entry = MockDbEntry(None,None,np.array([]))
        while num_msgs<max_msgs:
            num_msgs+=1
            msg = self.msgs.get()
            if isinstance(msg, MockDbEntry):
                mock_entry = msg
            if len(mock_entry.data)==len(src_data) and self.msgs.empty():
                break
        #print(self.msgs.get())
        np.testing.assert_array_equal(src_data, mock_entry.data)
        # self.assertEqual(len(mock_entry.intervals), 3)
        self.stop_server()

    @unittest.skip('fakejoule does not handle time bound queries so this test is not useful')
    def test_copies_new_data(self):
        server = FakeJoule()
        # create the source and destination streams
        src_data = create_source_data(server)  # helpers.create_data(src.layout)
        # dest has half the data
        dest = DataStream(id=1, name="dest", keep_us=100,
                          datatype=DataStream.DATATYPE.FLOAT32, updated_at=datetime.now(timezone.utc))
        dest.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in
                         range(3)]
        # destination is missing first interval but this won't be copied with the --new flag
        dest_interval = server.streams['/test/source'].intervals[1]
        dest_data = np.copy(src_data[dest_interval[0]:dest_interval[1]])
        server.add_stream('/test/destination', dest,
                          StreamInfo(int(dest_interval[0]), int(dest_interval[1]), len(dest_data)),
                          dest_data, [dest_interval])
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'copy', '--new', '/test/source', '/test/destination'])
        print(result.output)
        _print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        while self.msgs.empty():
            time.sleep(0.1)
        while not self.msgs.empty():
            msg = self.msgs.get()
            if type(msg) is MockDbEntry:
                print(msg)
        self.stop_server()

    def test_does_not_copy_existing_data(self):
        server = FakeJoule()
        # create the source and destination streams
        src_data = create_source_data(server)  # helpers.create_data(src.layout)
        # dest has the same intervals as source so nothing is copied
        ts = src_data['timestamp']
        intervals = server.streams['/test/source'].intervals

        dest = DataStream(id=1, name="dest", keep_us=100,
                          datatype=DataStream.DATATYPE.FLOAT32, updated_at=datetime.now(timezone.utc))
        dest.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in
                         range(3)]
        server.add_stream('/test/destination', dest, StreamInfo(int(ts[0]), int(ts[-1]),
                                                                len(ts)), src_data, intervals)
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'copy',
                                      '--start', str(ts[0]), '--end', str(ts[-1]),
                                      '/test/source', '/test/destination'])
        _print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        # only the annotations get was called (twice for each interval: src and dest)
        self.assertTrue(self.msgs.qsize(), len(intervals) * 2)
        self.stop_server()

    def test_creates_stream_if_necessary(self):
        server = FakeJoule()
        # create the source stream
        src = DataStream(id=0, name="source", keep_us=100,
                         datatype=DataStream.DATATYPE.FLOAT32,
                         updated_at=datetime.now(timezone.utc))
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]

        # source has 100 rows of data between [0, 100]
        src_data = helpers.create_data(src.layout, length=4)
        src_info = StreamInfo(int(src_data['timestamp'][0]), int(src_data['timestamp'][-1]), len(src_data))
        server.add_stream('/test/source', src, src_info, src_data, [[src_info.start, src_info.end]])
        server.add_stream('/test/source', src, src_info, src_data, [[src_info.start, src_info.end]])

        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'copy', '/test/source', '/test/destination'])
        _print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        while self.msgs.empty():
            time.sleep(0.1)
            print("waiting...")
        mock_entry = self.msgs.get()
        np.testing.assert_array_equal(src_data, mock_entry.data)
        self.stop_server()

    def test_when_source_is_empty(self):
        server = FakeJoule()
        # source has no data
        src_info = StreamInfo(None, None, 0)
        # create the source stream
        src = DataStream(id=0, name="source", keep_us=100,
                         datatype=DataStream.DATATYPE.FLOAT32, updated_at=datetime.now(timezone.utc))
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]
        server.add_stream('/test/source', src, src_info, np.ndarray([]))
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'copy', '/test/source', '/test/destination'])
        self.assertTrue('Error' in result.output and 'source' in result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()

    def test_when_destination_is_invalid(self):
        server = FakeJoule()
        # create the source stream
        src = DataStream(id=0, name="source", keep_us=100,
                         datatype=DataStream.DATATYPE.FLOAT32, updated_at=datetime.now(timezone.utc))
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]

        # source has 100 rows of data between [0, 100]
        src_data = helpers.create_data(src.layout, length=4)
        src_info = StreamInfo(int(src_data['timestamp'][0]), int(src_data['timestamp'][-1]), len(src_data))

        server.add_stream('/test/source', src, src_info, np.ndarray([]))
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'copy', '/test/source', 'badpath'])
        self.assertTrue('Error' in result.output and 'destination' in result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()

    def test_when_server_returns_invalid_data(self):
        server = FakeJoule()
        create_source_data(server)

        server.response = "notjson"
        server.http_code = 200
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'copy', '/test/source', '/stub/response'])
        self.assertIn('Error', result.output)

        self.assertEqual(result.exit_code, 1)
        self.stop_server()

    def test_incompatible_layouts(self):
        server = FakeJoule()
        create_source_data(server)
        dest = DataStream(id=1, name="dest", keep_us=100,
                          datatype=DataStream.DATATYPE.FLOAT32, updated_at=datetime.now(timezone.utc))
        dest.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in
                         range(5)]
        server.add_stream('/test/destination', dest, StreamInfo(None, None, 0), None)
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'copy', '/test/source', '/test/destination'])
        self.assertIn('not compatible', result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()

    def test_warn_on_different_elements(self):
        server = FakeJoule()
        create_source_data(server)
        dest = DataStream(id=1, name="dest", keep_us=100,
                          datatype=DataStream.DATATYPE.FLOAT32, updated_at=datetime.now(timezone.utc))
        dest.elements = [
            Element(name="different%d" % x, index=x, units='other', display_type=Element.DISPLAYTYPE.CONTINUOUS) for x
            in
            range(3)]
        server.add_stream('/test/destination', dest, StreamInfo(None, None, 0), None)
        self.start_server(server)
        runner = CliRunner()
        # does not copy without confirmation
        runner.invoke(main, ['data', 'copy', '/test/source', '/test/destination'])
        self.assertTrue(self.msgs.empty())
        # copies with confirmation
        result = runner.invoke(main, ['data', 'copy', '/test/source', '/test/destination'],
                               input='y\n')
        mock_entry = self.msgs.get()
        self.assertGreater(len(mock_entry.data), 0)
        self.assertEqual(result.exit_code, 0)
        self.stop_server()

    def test_start_must_be_before_end(self):
        server = FakeJoule()
        create_source_data(server)
        self.start_server(server)
        runner = CliRunner()
        # does not copy without confirmation
        result = runner.invoke(main, ['data', 'copy',
                                      '/test/source', '/test/destination',
                                      '--start', '1 hour ago', '--end', '2 hours ago'])
        self.assertEqual(result.exit_code, 1)
        self.assertIn('start', result.output)
        self.assertIn('end', result.output)
        self.stop_server()

    def test_data_read_errors(self):
        server = FakeJoule()
        server.stub_data_read = True
        server.http_code = 400
        server.response = 'joule error'
        create_source_data(server)
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'copy',
                                      '/test/source', '/test/destination'])
        self.assertEqual(result.exit_code, 1)
        self.assertTrue('Error' in result.output and 'source' in result.output)
        self.stop_server()

    def test_data_write_errors(self):
        server = FakeJoule()
        server.stub_data_write = True
        server.http_code = 400
        server.response = 'joule error'
        create_source_data(server)
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'copy',
                                      '/test/source', '/test/destination'])
        self.assertEqual(result.exit_code, 1)
        self.assertTrue('Error' in result.output and 'destination' in result.output)
        self.stop_server()


def _print_result_on_error(result):
    if result.exit_code != 0:
        print("output: ", result.output)
        print("exception: ", result.exception)


def create_source_data(server, break_into_intervals=True):
    # create the source stream
    src = DataStream(id=0, name="source", keep_us=100,
                     datatype=DataStream.DATATYPE.FLOAT32, updated_at=datetime.now(timezone.utc))
    src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]

    # source has 100 rows of data in four intervals between [0, 100]
    src_data = helpers.create_data(src.layout, length=100, start=0, step=1)
    if break_into_intervals:
        # insert the intervals
        pipe_data = np.hstack((src_data[:25],
                               pipes.interval_token(src.layout),
                               src_data[25:50],
                               pipes.interval_token(src.layout),
                               src_data[50:75],
                               pipes.interval_token(src.layout),
                               src_data[75:]))
        ts = src_data['timestamp']
        intervals = [[ts[0], ts[24]],
                     [ts[25], ts[49]],
                     [ts[50], ts[74]],
                     [ts[75], ts[99]]]
    else:
        pipe_data = src_data
        intervals = [[src_data['timestamp'][0],src_data['timestamp'][-1]]]
    src_info = StreamInfo(int(src_data['timestamp'][0]), int(src_data['timestamp'][-1]), len(src_data))
    server.add_stream('/test/source', src, src_info, pipe_data, intervals)
    return src_data



import click

from joule.cli.lazy_group import LazyGroup

@click.group(name="event",
             cls=LazyGroup,
             lazy_subcommands={
                 "info": "joule.cli.event.info.cli_info",
                 "move": "joule.cli.event.move.cli_move",
                 "rename": "joule.cli.event.rename.cli_rename",
                 "copy": "joule.cli.event.copy.cli_copy"
             })
def events():
    """Manage event streams."""
    pass  




import unittest
from datetime import datetime, timezone
from joule.models import (DataStream, Element, data_stream)
from joule.models.data_stream import from_json as data_stream_from_json
from joule.errors import ConfigurationError
from tests import helpers


class TestStream(unittest.TestCase):
    def setUp(self):
        us_in_week = int(7 * 24 * 60 * 60 * 1e6)
        self.my_stream = data_stream.DataStream(name="test",
                                                description="test_description",
                                                datatype=DataStream.DATATYPE.FLOAT32,
                                                keep_us=us_in_week,
                                                decimate=True,
                                                updated_at=datetime.now(timezone.utc))
        self.my_stream.elements.append(Element(name="e1", index=0,
                                               display_type=Element.DISPLAYTYPE.CONTINUOUS))
        self.base_config = helpers.parse_configs(
            """[Main]
                 name = test
                 description = test_description
                 path = /some/path/to/data
                 datatype = float32
                 keep = 1w
                 decimate = yes
               [Element1]
                 name = e1
            """)

    def test_attributes(self):
        my_stream = DataStream(name='test', description='a test',
                               datatype=DataStream.DATATYPE.FLOAT32, keep_us=DataStream.KEEP_ALL,
                               updated_at=datetime.now(timezone.utc))
        self.assertIsNotNone(my_stream)
        # has a meaningful string representation
        self.assertIn("test", "%r" % my_stream)
        # if remote is not set, the pseudo attributes are empty
        self.assertEqual(my_stream.remote_node, '')
        self.assertEqual(my_stream.remote_path, '')
        # if remote is set, these attributes have values
        my_stream.set_remote("remote_url", "remote_path")
        self.assertEqual(my_stream.remote_node, "remote_url")
        self.assertEqual(my_stream.remote_path, "remote_path")

    def test_json(self):
        my_stream = DataStream(id=0, name='test', decimate=True,
                               datatype=DataStream.DATATYPE.INT16,
                               updated_at=datetime.now(timezone.utc))
        for j in range(4):
            my_stream.elements.append(Element(name="e%d" % j, index=j,
                                              display_type=Element.DISPLAYTYPE.CONTINUOUS))
        # turns streams into json
        json = my_stream.to_json()
        self.assertTrue(json['decimate'])
        self.assertEqual(json['name'], 'test')
        self.assertEqual(len(json['elements']), 4)
        # builds streams from json
        new_stream = data_stream.from_json(json)
        self.assertEqual(new_stream.id, my_stream.id)
        self.assertEqual(len(new_stream.elements), len(my_stream.elements))

    def test_parses_config(self):
        s = data_stream.from_config(self.base_config)
        self.assertEqual(s.name, self.base_config["Main"]["name"])

    def test_allows_no_keep(self):
        config = helpers.parse_configs(
            """[Main]
                 name = test
                 description = test_description                 
                 path = /simple/demo
                 datatype = float32
                 keep = None
               [Element1]
                 name=1
            """)
        s = data_stream.from_config(config)
        self.assertEqual(s.keep_us, 0)

    def test_allows_long_keeps(self):
        config = helpers.parse_configs(
            """[Main]
                 name = test
                 description = test_description                 
                 path = /simple/demo
                 datatype = float32
                 keep = 100h
               [Element1]
                 name=1
            """)
        s = data_stream.from_config(config)
        self.assertEqual(s.keep_us, 100 * 60 * 60 * 1e6)

    def test_allows_no_description(self):
        self.base_config.remove_option("Main", "description")
        parsed_stream = data_stream.from_config(self.base_config)
        self.assertEqual(parsed_stream.description, "")

    def test_computes_layout(self):
        """data_format returns float32_4 for example"""
        self.my_stream.elements = []
        for i in range(4):
            self.my_stream.elements.append(Element(name="%d" % i))
        self.assertEqual(self.my_stream.layout, "float32_4")
        self.assertEqual(self.my_stream.data_width, 5)

    def test_has_string_representation(self):
        self.assertRegex("%s" % self.my_stream, self.my_stream.name)

    def test_updates(self):
        self.my_stream.update_attributes({
            "name": "new name",
            "description": "new description",
            "elements": [
                {"name": "new name"},
            ]
        })
        self.assertEqual("new name", self.my_stream.name)
        self.assertEqual("new description", self.my_stream.description)
        # updates element attributes
        self.assertEqual("new name", self.my_stream.elements[0].name)

        # number of elements in the request must match the elements in the stream
        with self.assertRaises(ConfigurationError):
            self.my_stream.update_attributes({
                "name": "new name",
                "description": "new description",
                "elements": [
                    {"name": "new name"},
                    {"extra", "causes error"}
                ]
            })

    def test_merge_configs(self):
        stream_copy = data_stream_from_json(self.my_stream.to_json())
        # when there is no change, updated_at is the same, method returns false
        prev_update = self.my_stream.updated_at
        self.assertFalse(self.my_stream.merge_configs(stream_copy))
        self.assertEqual(prev_update, self.my_stream.updated_at)
        # when there is a change, it is recorded and the updated_at is refreshed
        stream_copy.keep_us = self.my_stream.keep_us + 1
        self.assertTrue(self.my_stream.merge_configs(stream_copy))
        self.assertEqual(self.my_stream.keep_us, stream_copy.keep_us)
        self.assertGreater(self.my_stream.updated_at, prev_update)

    


from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web
from sqlalchemy.orm import Session
import asyncio
from unittest import mock
from joule.api.node.node_info import NodeInfo
from joule.models import Follower
import joule.controllers
from tests.controllers.helpers import create_db, MockStore
from joule import app_keys
from joule.constants import EndPoints

import testing.postgresql
psql_key = web.AppKey("psql", testing.postgresql.Postgresql)

class TestFollowerController(AioHTTPTestCase):

    async def tearDownAsync(self):
        self.app[app_keys.db].close()
        await asyncio.to_thread(self.app[psql_key].stop)
        await self.client.close()

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        app[app_keys.name] = "test-node"
        app[app_keys.cafile] = "stubbed_cafile"
        app[app_keys.db], app[psql_key] = await asyncio.to_thread(lambda: create_db([]))
        return app

    @mock.patch('joule.controllers.follower_controller.TcpNode')
    async def test_adds_follower_node(self, mock_node_class):
        mock_node = mock.AsyncMock()
        mock_node.info.return_value = NodeInfo(version="1.0",name="test-node-follower",path="/path",
                                               size_reserved=0,size_other=0,size_free=0,size_db=0)
        mock_node_class.return_value = mock_node
        # must send a JSON request
        resp = await self.client.post(EndPoints.follower, 
                                      params={"key":"key",
                                            "port":8080,
                                            "scheme":"http",
                                            "base_uri":"/endpoint"})
        self.assertEqual(resp.status, 400)
        self.assertIn('content-type', await resp.text())

        # must have all parameters in request
        resp = await self.client.post(EndPoints.follower, 
                                      json={"key":"key",
                                            "port":'8080',
                                            "scheme":"http"})
        self.assertEqual(resp.status, 400)
        self.assertIn('base_uri', await resp.text())

        # handles ValueErrors (not clear why this could happen- maybe an endpoint returns invalid JSON by coincidence?)
        # raise a ValueError on the info call
        mock_node.info.side_effect = ValueError("bad value")
        resp = await self.client.post(EndPoints.follower, 
                                      json={"key":"key",
                                            "port":'8080',
                                            "scheme":"http",
                                            "base_uri":"/endpoint"})
        self.assertEqual(resp.status, 400)

    async def test_adds_follower_node_network_errors(self):
        # parameters must have valid values
        resp = await self.client.post(EndPoints.follower, 
                                      json={"key":"key",
                                            "port":"9999",
                                            "scheme":"invalid",
                                            "base_uri":"/endpoint"})
        self.assertEqual(resp.status, 400)
        self.assertIn('no route', await resp.text())

    async def test_deletes_follower_nodes(self):
        db = self.app[app_keys.db]
        db.add(Follower(name="test-follower",key="key",location="http://othernode.local:8080"))
        self.assertEqual(db.query(Follower).count(),1)

        # name must be correct
        resp = await self.client.delete(EndPoints.follower,params={"name":"not-present"})
        self.assertEqual(resp.status, 404)
        # the follower should not be removed
        self.assertEqual(db.query(Follower).count(),1)

        # name must be specified
        resp = await self.client.delete(EndPoints.follower)
        self.assertEqual(resp.status, 400)
        # the follower should not be removed
        self.assertEqual(db.query(Follower).count(),1)



import time
import signal
import sys


def main():
    while(True):
        time.sleep(0.1)
        print("in run_forever")
        sys.stdout.flush()


def handler(signum, frame):
    print("caught %d, ignoring" % signum)
    sys.stdout.flush()


if __name__ == "__main__":
    print("running forever!!!\n")
    signal.signal(signal.SIGTERM, handler)
    main()




from click.testing import CliRunner

import warnings
import numpy as np
import logging
import asyncio
import tempfile
import h5py
import unittest
from datetime import datetime, timezone

from ..fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main
from joule.models import DataStream, Element, StreamInfo, pipes
from tests import helpers

warnings.simplefilter('always')
log = logging.getLogger('aiohttp.access')
log.setLevel(logging.WARNING)


class TestDataRead(FakeJouleTestCase):

    def test_reads_data(self):
        server = FakeJoule()
        # create the source stream
        src = DataStream(id=0, name="source", keep_us=100,
                         datatype=DataStream.DATATYPE.FLOAT32,
                         updated_at=datetime.now(timezone.utc))
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]
        # source has 100 rows of data between [0, 100]
        src_data = helpers.create_data(src.layout)
        src_info = StreamInfo(int(src_data['timestamp'][0]), int(src_data['timestamp'][-1]), len(src_data))
        server.add_stream('/test/source', src, src_info, src_data)

        self.start_server(server)
        runner = CliRunner()
        # add in some extra parameters to make sure they are parsed
        result = runner.invoke(main, ['data', 'read', '/test/source',
                                      '--start', '0', '--end', '1 hour ago'])
        _print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        output = result.output.split('\n')
        for x in range(len(src_data)):
            row = src_data[x]
            expected = "%d %s" % (row['timestamp'], ' '.join('%f' % x for x in row['data']))
            self.assertIn(expected, output[x + 1])

        self.stop_server()

    def test_reads_decimated_data(self):

        server = FakeJoule()
        # create the source stream
        src = DataStream(id=0, name="source", keep_us=100, datatype=DataStream.DATATYPE.FLOAT32,
                         updated_at=datetime.now(timezone.utc))
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]
        # source has 200 rows of data between [0, 200] in two intervals
        src_data = np.hstack((helpers.create_data(src.decimated_layout, start=0, length=100, step=1),
                              pipes.interval_token(src.decimated_layout),
                              helpers.create_data(src.decimated_layout, start=100, length=100, step=1)))

        src_info = StreamInfo(int(src_data['timestamp'][0]), int(src_data['timestamp'][-1]), len(src_data))
        server.add_stream('/test/source', src, src_info, src_data)
        self.start_server(server)

        # mark the intervals and show the bounds
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'read', '/test/source',
                                      '--start', '0', '--end', '1 hour ago',
                                      '--max-rows', '28', '--mark-intervals',
                                      '--show-bounds'])
        _print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        output = result.output.split('\n')
        for x in range(len(src_data)):
            row = src_data[x]
            if row == pipes.interval_token(src.decimated_layout):
                expected = '# interval break'
            else:
                expected = "%d %s" % (row['timestamp'], ' '.join('%f' % x for x in row['data']))
            # import pdb; pdb.set_trace()
            self.assertIn(expected, output[x + 1],msg=f"{output[x+2]}!={expected}")


        # do not mark the intervals and hide the bounds
        runner = CliRunner()

        result = runner.invoke(main, ['data', 'read', '/test/source',
                                      '--start', '0', '--end', '1 hour ago',
                                      '--max-rows', '28'])
        _print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        output = result.output.split('\n')
        offset = 0
        for x in range(len(src_data)):
            row = src_data[x]
            if row == pipes.interval_token(src.decimated_layout):
                offset = 1
                continue
            else:
                expected = "%d %s" % (row['timestamp'], ' '.join('%f' % x for x in row['data'][:3]))
            if expected not in output[x - offset + 1]:
                print(f"expected: {expected}")
                print(f"output: {output[x - offset + 1]}")
                breakpoint()
                print("hmmm")
                
            self.assertIn(expected, output[x - offset + 1])

        self.stop_server()

    def test_reads_selected_elements_of_decimated_data(self):
        server = FakeJoule()
        # create the source stream
        src = DataStream(id=0, name="source", keep_us=100, datatype=DataStream.DATATYPE.FLOAT32,
                         updated_at=datetime.now(timezone.utc))
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]
        # source has 200 rows of data between [0, 200] in two intervals
        src_data = np.hstack((helpers.create_data(src.decimated_layout, start=0, length=100, step=1),
                              pipes.interval_token(src.decimated_layout),
                              helpers.create_data(src.decimated_layout, start=100, length=100, step=1)))

        src_info = StreamInfo(int(src_data['timestamp'][0]), int(src_data['timestamp'][-1]), len(src_data))
        server.add_stream('/test/source', src, src_info, src_data)
        self.start_server(server)

        # mark the intervals and show the bounds
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'read', '/test/source',
                                      '--start', '0', '--end', '1 hour ago',
                                      '--max-rows', '28', '--mark-intervals',
                                      '--show-bounds', '-i', '0,2'])
        _print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        output = result.output.split('\n')
        for x in range(len(src_data)):
            row = src_data[x]
            if row == pipes.interval_token(src.decimated_layout):
                expected = '# interval break'
            else:
                data = row['data'][[0, 2, 3, 5, 6, 8]]
                expected = "%d %s" % (row['timestamp'], ' '.join('%f' % x for x in data))
            self.assertIn(expected, output[x + 1])

        # do not mark the intervals and hide the bounds
        runner = CliRunner()

        result = runner.invoke(main, ['data', 'read', '/test/source',
                                      '--start', '0', '--end', '1 hour ago',
                                      '--max-rows', '28'])
        _print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        output = result.output.split('\n')
        offset = 0
        for x in range(len(src_data)):
            row = src_data[x]
            if row == pipes.interval_token(src.decimated_layout):
                offset = 1
                continue
            else:
                expected = "%d %s" % (row['timestamp'], ' '.join('%f' % x for x in row['data'][:3]))
            self.assertIn(expected, output[x - offset + 1])

        self.stop_server()

    def test_reads_selected_elements(self):
        server = FakeJoule()
        # create the source stream
        src = DataStream(id=0, name="source", keep_us=100, datatype=DataStream.DATATYPE.FLOAT32,
                         updated_at=datetime.now(timezone.utc))
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]
        # source has 100 rows of data between [0, 100]
        src_data = helpers.create_data(src.layout)
        src_info = StreamInfo(int(src_data['timestamp'][0]), int(src_data['timestamp'][-1]), len(src_data))
        server.add_stream('/test/source', src, src_info, src_data)

        self.start_server(server)
        runner = CliRunner()
        # add in some extra parameters to make sure they are parsed
        result = runner.invoke(main, ['data', 'read', '/test/source', '-i', '1,0',
                                      '--start', '0', '--end', '1 hour ago'])
        _print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        output = result.output.split('\n')
        for x in range(len(src_data)):
            row = src_data[x]
            expected = "%d %s" % (row['timestamp'], ' '.join('%f' % x for x in row['data'][[1, 0]]))
            self.assertIn(expected, output[x + 1])

        self.stop_server()

    def test_reads_selected_elements_to_file(self):
        server = FakeJoule()
        # create the source stream
        src = DataStream(id=0, name="source", keep_us=100, datatype=DataStream.DATATYPE.INT16,
                         updated_at=datetime.now(timezone.utc))
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]
        # source has 100 rows of data between [0, 100]
        src_data = helpers.create_data(src.layout)
        src_info = StreamInfo(int(src_data['timestamp'][0]), int(src_data['timestamp'][-1]), len(src_data))
        server.add_stream('/test/source', src, src_info, src_data)

        self.start_server(server)
        runner = CliRunner()
        # add in some extra parameters to make sure they are parsed
        with tempfile.NamedTemporaryFile() as data_file:
            result = runner.invoke(main, ['data', 'read', '/test/source',
                                          '--start', '0', '--end', '1 hour ago',
                                          '-i', '0,2',
                                          '--file', data_file.name])

            _print_result_on_error(result)
            self.assertEqual(result.exit_code, 0)
            h5_file = h5py.File(data_file.name, 'r')
            self.assertEqual(src_data['data'].dtype, h5_file['data'].dtype)
            self.assertEqual(h5_file['timestamp'].dtype, np.dtype('i8'))

            np.testing.assert_array_almost_equal(h5_file['data'], src_data['data'][:, [0, 2]])
            np.testing.assert_array_almost_equal(h5_file['timestamp'], src_data['timestamp'][:, None])

            h5_file.close()

        self.stop_server()

    def test_reads_data_to_file(self):
        server = FakeJoule()
        # create the source stream
        src = DataStream(id=0, name="source", keep_us=100, datatype=DataStream.DATATYPE.FLOAT32,
                         updated_at=datetime.now(timezone.utc))
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]
        # source has 100 rows of data between [0, 100]
        src_data = helpers.create_data(src.layout)
        src_info = StreamInfo(int(src_data['timestamp'][0]), int(src_data['timestamp'][-1]), len(src_data))
        server.add_stream('/test/source', src, src_info, src_data)

        self.start_server(server)
        runner = CliRunner()
        # add in some extra parameters to make sure they are parsed
        with tempfile.NamedTemporaryFile() as data_file:
            result = runner.invoke(main, ['data', 'read', '/test/source',
                                          '--start', '0', '--end', '1 hour ago',
                                          '--file', data_file.name])

            _print_result_on_error(result)
            self.assertEqual(result.exit_code, 0)
            h5_file = h5py.File(data_file.name, 'r')
            np.testing.assert_array_almost_equal(src_data['timestamp'][:, None], h5_file['timestamp'])
            np.testing.assert_array_equal(src_data['data'], h5_file['data'])

            h5_file.close()

        self.stop_server()

    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        # create the source stream
        src = DataStream(id=0, name="source", keep_us=100, datatype=DataStream.DATATYPE.FLOAT32,
                         updated_at=datetime.now(timezone.utc))
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]
        # source has 200 rows of data between [0, 200] in two intervals
        src_data = np.hstack((helpers.create_data(src.decimated_layout, start=0, length=100, step=1),
                              pipes.interval_token(src.decimated_layout),
                              helpers.create_data(src.decimated_layout, start=100, length=100, step=1)))

        src_info = StreamInfo(int(src_data['timestamp'][0]), int(src_data['timestamp'][-1]), len(src_data))
        server.add_stream('/test/source', src, src_info, src_data)

        server.response = "test error"
        server.http_code = 500
        server.stub_data_read = True
        self.start_server(server)
        runner = CliRunner()

        with self.assertLogs(level=logging.ERROR):
            runner.invoke(main, ['data', 'read', '/test/source', '--start', 'now'])

        self.stop_server()

    def test_handles_bad_parameters(self):
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'read', '/test/source', '--start', 'invalid'])
        self.assertIn('start time', result.output)
        self.assertEqual(result.exit_code, 1)
        result = runner.invoke(main, ['data', 'read', '/test/source', '--end', 'invalid'])
        self.assertIn('end time', result.output)
        self.assertEqual(result.exit_code, 1)


def _print_result_on_error(result):
    if result.exit_code != 0:
        print("output: ", result.output)
        print("exception: ", result.exception)



from click.testing import CliRunner
import os
import logging

import warnings
from tests.cli.fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main


FOLDER_LIST = os.path.join(os.path.dirname(__file__), 'folders.json')
warnings.simplefilter('always')
aio_log = logging.getLogger('aiohttp.access')
aio_log.setLevel(logging.WARNING)


class TestFolderList(FakeJouleTestCase):

    def test_lists_folders(self):
        server = FakeJoule()
        with open(FOLDER_LIST, 'r') as f:
            server.response = f.read()
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['folder', 'list'])
        self.assertEqual(result.exit_code, 0)
        output = result.output
        # make sure the folders are listed
        for folder in ['basic', 'aux', 'event', 'sensors']:
            self.assertIn(folder, output)
        # make sure the data streams are listed
        for stream in ['Accel', 'Encoder', 'Gyro']:
            self.assertIn(stream, output)
        # make sure the event streams are listed
        for stream in ['events0', 'events1', 'events2']:
            self.assertIn(stream, output)
        # should  check for layout strings
        self.assertNotIn("float32_3", output)
        self.stop_server()

    def test_lists_streams_with_options(self):
        server = FakeJoule()
        with open(FOLDER_LIST, 'r') as f:
            server.response = f.read()
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['folder', 'list', '-s', '-l'])
        self.assertEqual(result.exit_code, 0)
        output = result.output
        # check for the augmented legend
        self.assertIn("configured", output)
        # check for layout strings
        self.assertIn("float32_3", output)
        self.stop_server()

    def test_when_server_returns_invalid_data(self):
        server = FakeJoule()
        server.response = "notjson"
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['folder', 'list'])
        self.assertIn('Error', result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()

    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        error_msg = "test error"
        error_code = 500
        server.response = error_msg
        server.http_code = error_code
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['folder', 'list'])
        self.assertIn('%d' % error_code, result.output)
        self.assertIn(error_msg, result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()



from click.testing import CliRunner
import os
import warnings
import asyncio

from ..fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main

STREAM_INFO = os.path.join(os.path.dirname(__file__), 'stream.json')
warnings.simplefilter('always')


class TestFolderDelete(FakeJouleTestCase):
    def test_deletes_folder(self):
        server = FakeJoule()
        self.start_server(server)
        runner = CliRunner()
        # deletes folders
        result = runner.invoke(main,
                               ['folder', 'delete',
                                '/the/folder'])

        self.assertEqual(result.exit_code, 0)
        params = self.msgs.get()
        self.assertEqual(params['path'], '/the/folder')



        # recursively deletes folders
        result = runner.invoke(main,
                               ['folder', 'delete',
                                '/the/folder', '-r'],
                               input='y\n')
        self.assertEqual(result.exit_code, 0)
        params = self.msgs.get()
        self.assertEqual(params['path'], '/the/folder')
        self.assertIn('recursive', params)

        self.stop_server()

    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        error_msg = "test error"
        error_code = 500
        server.response = error_msg
        server.http_code = error_code
        server.stub_folder_destroy = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['folder', 'delete', '/the/folder'])
        self.assertIn('%d' % error_code, result.output)
        self.assertIn(error_msg, result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()



#!/usr/bin/env python3

import joule.client
import joule.utilities
import numpy as np
import asyncio
import sys

rows = 100
freq = 40  # Hz


class RepeatInserter(joule.client.ReaderModule):

    async def run(self, parsed_args, output):
        data = 100 * np.sin(np.arange(0, 2 * np.pi, 2 * np.pi / rows))
        data.shape = (rows, 1)

        # 1540239607500500
        data_ts = joule.utilities.time_now()
        orig_ts = data_ts
        count = 0
        ts_inc = 1 / rows * (1 / freq) * 1e6  # microseconds
        while not self.stop_requested:
            top_ts = data_ts + 100 * ts_inc
            count += 1
            ts = np.array(np.linspace(data_ts, top_ts, rows,
                                      endpoint=False), dtype=np.int64)
            ts.shape = (rows, 1)
            ts_data = np.hstack((ts, data))
            await output.write(ts_data)
            data_ts = top_ts
            await asyncio.sleep(1 / freq)
            if count==freq*10 and False:
                # cause overlapping timestamp
                data_ts = orig_ts
                print("resetting", file=sys.stderr)
                count = 0

if __name__ == "__main__":
    r = RepeatInserter()
    r.start()



from aiohttp.test_utils import AioHTTPTestCase
from aiohttp import web
import aiohttp
import json
import asyncio
import joule.controllers
from tests.controllers.helpers import create_db, MockSupervisor, MockWorker
from joule.models.proxy import Proxy
from joule.api.event_stream import EventStream as ApiEventStream, from_json as event_stream_from_json
from joule.api.event import Event
from joule import app_keys
from joule.constants import EndPoints

import testing.postgresql
psql_key = web.AppKey("psql", testing.postgresql.Postgresql)

class TestAppController(AioHTTPTestCase):

    async def tearDownAsync(self):
        self.app[app_keys.db].close()
        await asyncio.to_thread(self.app[psql_key].stop)
        await self.client.close()

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        app[app_keys.db], app[psql_key] = await asyncio.to_thread(lambda: create_db([],[]))
        self.supervisor = MockSupervisor()
        self.supervisor.workers = [MockWorker('worker1', {}, {}, 10, is_app=True),
                                   MockWorker('worker2', {}, {}, 20, is_app=False)]
        self.supervisor.proxies = [Proxy('proxy1', 11, 'http://localhost:8080'),
                                   Proxy('proxy2', 22, 'http://localhost:8081')]
        app[app_keys.supervisor] = self.supervisor
        return app
    
    async def test_get_apps(self):
        resp = await self.client.get(EndPoints.app_json)
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        # make sure the data app worker is in the response
        self.assertIn({'id': 'm10', 'name': 'worker1'}, data)
        # make suer the non-app worker is not in the response
        self.assertNotIn({'id': 'm20', 'name': 'worker2'}, data)
        # make sure both proxies are in the response
        self.assertIn({'id': 'p11', 'name': 'proxy1'}, data)
        self.assertIn({'id': 'p22', 'name': 'proxy2'}, data)

    async def test_get_app_auth(self):
        resp = await self.client.get(EndPoints.app_auth, headers={"X-App-Id": "m10"})
        self.assertEqual(resp.status, 200)
        # mock supervisor return --uuid-- for the module socket
        self.assertEqual(resp.headers['X-Proxy-Path'], "http://unix:--10--:/")

        resp = await self.client.get(EndPoints.app_auth, headers={"X-App-Id": "p11"})
        self.assertEqual(resp.status, 200)
        # mock supervisor returns the url for the proxy
        self.assertEqual(resp.headers['X-Proxy-Path'], "http://localhost:8080")


import unittest

import joule.api
import subprocess
from joule import api
import asyncio

class TestBasicAPI(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.node1 = api.get_node("node1.joule")
        followers = await self.node1.follower_list()
        self.assertEqual(len(followers), 1)
        self.node2 = followers[0]
        # add the root user as a follower to node2.joule
        user = await self.node2.master_add("user", "cli")
        # follower uses built-in HTTP server on port 8080
        proc = await asyncio.create_subprocess_shell(f"coverage run --rcfile=/joule/.coveragerc -m joule.cli node add node2.joule http://node2.joule:8080 {user.key}")
        await proc.communicate()
        """
        node1.joule:
        /main/folder/added:int32[x]
        node2.joule:
        /main/folder/base:int32[x]
        """

    async def asyncTearDown(self):
        await self.node2.master_delete("user", "cli")
        await self.node1.close()
        await self.node2.close()

    async def test_retrieves_streams(self):
        logs = await self.node1.module_logs("Remote")
        self.assertGreater(len(logs), 0)
        self.assertIn("starting module", logs[0])
        added_stream = await self.node1.data_stream_get("/main/folder/added")
        base_stream = await self.node2.data_stream_get("/main/folder/base")
        # make sure added_stream has data
        added_data = await self.node1.data_subscribe(added_stream)
        base_data = await self.node2.data_subscribe(base_stream)
        base_block = await base_data.read()
        added_block = await added_data.read()
        # align the time stamps
        diff = base_block['timestamp'][0] - added_block['timestamp'][0]
        print("the diff is %d" % diff)
        await added_data.close()
        await base_data.close()

    async def test_copies_events(self):
        events = []
        for i in range(1000, 2000, 100):
            events.append(joule.api.Event(start_time=i, end_time=i + 50,
                                          content={'name': f"Event{i}"}))
        event_stream = joule.api.EventStream("TestEvents")
        event_stream = await self.node1.event_stream_create(event_stream, "/eventstreams")
        await self.node1.event_stream_write(event_stream, events)
        cmd = "coverage run --rcfile=/joule/.coveragerc -m joule.cli event copy /eventstreams/TestEvents /eventstreams/TestEventsCopy -d node2.joule"
        proc = await asyncio.create_subprocess_shell(cmd)
        await proc.communicate()
        copied_events = await self.node2.event_stream_read_list("/eventstreams/TestEventsCopy")
        self.assertListEqual(events, copied_events)



#!/usr/bin/env python3
"""
InStream1 ==> x2 ==> InStream2
InStream2 ==> x3 ==> OutStream2
"""


import joule.client
import argparse
import json

import adder

"""
run 1 or more adders
arguments:
  npaths = number of adders to run
  offsets = json array of offsets (len==npaths)

path1 --> + offset[0] --> path1
path2 --> + offset[1] --> path2
...more

"""


class CompositeAdder(joule.client.CompositeModule):
    " add offsets to input streams"

    def custom_args(self, parser):
        parser.add_argument("--npaths", type=int, required=True)
        parser.add_argument("--offsets",
                            required=True,
                            help="JSON array, length must equal npaths")
        
    async def setup(self, parsed_args, inputs, outputs):
        npaths = parsed_args.npaths
        offsets = json.loads(parsed_args.offsets)
        tasks = []
        for x in range(npaths):
            a = adder.Adder()
            args = argparse.Namespace(offset=offsets[x])
            t = a.run(args,
                      {"input": inputs["path%d" % (x+1)]},
                      {"output": outputs["path%d" % (x+1)]})
            tasks.append(t)
        return tasks

    
if __name__ == "__main__":
    r = CompositeAdder()
    r.start()



from aiohttp import web
from sqlalchemy.orm import Session, exc
import secrets
from typing import Dict

from joule.api.session import TcpSession
from joule.models.master import Master, make_key
from joule.utilities.misc import detect_url
from joule import errors
from joule import app_keys
from joule.controllers.helpers import read_json
from joule.constants import EndPoints

async def index(request: web.Request):
    db: Session = request.app[app_keys.db]
    masters = db.query(Master).all()
    return web.json_response([m.to_json() for m in masters])

async def add(request: web.Request):
    """
    Contact specified node and add self as a follower
    """
    db: Session = request.app[app_keys.db]
    local_name: str = request.app[app_keys.name]
    body = await read_json(request)
    try:
        master_type = body['master_type'].lower()
        if master_type not in ['user', 'joule', 'lumen']:
            raise web.HTTPBadRequest(reason="master_type must be [user|joule|lumen]")
        
        # Use the API token that authenticated this request to find the master who
        # is granting access to this user/node
        grantor_key = request.headers["X-API-KEY"]
        grantor_id = db.query(Master.id).filter(Master.key == grantor_key).one()[0]

        # How we add the master depends on the type, a user or a node (joule or lumen)
        if master_type == 'user':
            return _add_user(db, body, local_name, grantor_id)
        elif master_type == 'joule' or master_type == 'lumen':
            local_port: int = request.app[app_keys.port]
            local_scheme: str = request.app[app_keys.scheme]
            local_uri: str = request.app[app_keys.base_uri]
            cafile = request.app[app_keys.cafile]
            return await _add_device(db, master_type, body, local_name, local_port, local_scheme, local_uri, cafile, grantor_id)
        
    except (KeyError, ValueError) as e:
            raise web.HTTPBadRequest(reason="Invalid request %s" % str(e))
        
### BEGIN: Helper Functions for Add ###

def _add_user(db: Session, data: Dict, local_name: str, grantor_id) -> web.Response:
    # check if this name is associated with a master entry
    my_master = db.query(Master). \
        filter(Master.type == Master.TYPE.USER). \
        filter(Master.name == data['identifier']).first()
    if my_master is not None:
        raise web.HTTPBadRequest(reason="User already exists")
    # see if API key was included in request
    if 'api_key' in data and data['api_key'] is not None:
        key = data['api_key']
        if len(key) < 32:
            raise web.HTTPBadRequest(reason="API Key must be at least 32 characters")
    # otherwise make a new key
    else:
        key = make_key()
    my_master = Master(name=data['identifier'],
                        type=Master.TYPE.USER,
                        grantor_id=grantor_id,
                        key=key)
    db.add(my_master)
    db.commit()
    return web.json_response({"key": my_master.key,
                              "name": local_name})
    
    

async def _add_device(db: Session, master_type: str, data: Dict, local_name: str, port, scheme, uri, cafile, grantor_id) -> web.Response:
    try:
        # create a new node API key
        my_master = Master(key=make_key(), grantor_id=grantor_id)
        if master_type == 'joule':
            my_master.type = Master.TYPE.JOULE_NODE
        else:
            my_master.type = Master.TYPE.LUMEN_NODE
        # make a temporary name
        my_master.name = "pending_%s" % secrets.token_hex(5)
        db.add(my_master)
        db.commit()
        # post this key to the master
        # keep track of whether the operation is successful, don't fail on error
        # because it might not be possible to coordinate the correct URL's, let the user
        # fix this later through the Lumen web interface or some other way.
        success = False
        error_msg = ""
        if master_type == 'joule':
            coro = _send_node_key(my_master.key, data['identifier'], port,
                                    local_name, scheme, uri, cafile)
        else:
            coro = _send_lumen_key(my_master.key, data['identifier'], port,
                                    local_name, scheme, uri, cafile,
                                    data['lumen_params'])
        try:
            name = await coro
            success = True
        except errors.ApiError as e:
            error_msg = str(e)
            name = data['identifier']
            success = False

        # remove the node if it already exists, since we are creating a new API key
        # this will have the effect of expiring the old key
        db.query(Master). \
            filter(Master.type == my_master.type). \
            filter(Master.name == name).delete()
        my_master.name = name
        db.add(my_master)
        db.commit()
        if success:
            return web.json_response({"name": my_master.name})
        else:
            return web.json_response({"error": error_msg}, status=400) 
    except (ValueError, KeyError) as e:
        # Unlikely to happen, maybe if the request is malformed
        db.delete(my_master)
        db.commit()
        raise web.HTTPBadRequest(reason="Cannot add master: [%s]" % str(e))
  
async def _send_node_key(key: str,
                         identifier: str,
                         local_port: int,
                         local_name: str,
                         local_scheme: str,
                         local_uri: str,
                         cafile: str) -> str:
    # if the identifier is an IP address or a domain name, turn it into a URL
    if identifier.startswith("http://") or identifier.startswith("https://"):
        url = identifier
    else:
        url = await detect_url(identifier, 443)
        if url is None:
            raise errors.ApiError(f"cannot connect to {identifier}, try using a full URL like http(s)://...")
    
    ##NOTE: The key is not needed to access /followers but the parameter is required by TcpSession
    # This is the only route that is not authenticated by an API key
    session = TcpSession(url, key, cafile)
    params = {'key': key,
              'port': local_port,
              'name': local_name,
              'base_uri': local_uri,
              'scheme': local_scheme}
    if cafile != "":
        params["name_is_host"] = 1
    resp = await session.post(EndPoints.follower, json=params)
    await session.close()
    return resp['name']


async def _send_lumen_key(key: str,
                          identifier: str,
                          local_port: int,
                          local_name: str,
                          local_scheme: str,
                          local_uri: str,
                          cafile: str,
                          lumen_params: Dict) -> str:
    # if the identifier is an IP address or a domain name, turn it into a URL
    if not identifier.startswith("http"):
        url = await detect_url(identifier + "/lumen")
        if url is None:
            raise errors.ApiError(f"cannot connect to {identifier}, try using a full URL like http(s)://...")
    else:
        url = identifier

    # no key needed to access the lumen /nilms end point
    session = TcpSession(url, "stub_key", cafile)
    params = {'api_key': key,
              'port': local_port,
              'name': local_name,
              'base_uri': local_uri,
              'scheme': local_scheme}
    if lumen_params is not None:
        params = {**params, **lumen_params}
    # TODO enable this for CA checking on Lumen server
    # if cafile != "":
    #    params["name_is_host"] = 1
    try:
        await session.post("/nilms.json", json=params)
    finally:
        await session.close()
    return identifier

### END: Helper Functions for Add ###


async def delete(request: web.Request):
    """
    Remove the specified node from the list of masters
    """
    db: Session = request.app[app_keys.db]
    node_name: str = request.app[app_keys.name]
    # to eliminate uninitialized warnings
    str_master_type = ""
    name = ""
    try:
        name = request.query["name"]
        str_master_type = request.query["master_type"]
        if str_master_type != "user":
            str_master_type += "_node"
        master_type = Master.TYPE[str_master_type.upper()]
        master = db.query(Master). \
            filter(Master.name == name). \
            filter(Master.type == master_type). \
            one()
        if master.key == request.headers["X-API-KEY"]:
            raise web.HTTPBadRequest(reason="cannot delete yourself, this would lock you out of the node")

        db.delete(master)
    except (KeyError, ValueError):
        raise web.HTTPBadRequest(reason="specify name and master_type")
    except exc.NoResultFound:
        raise web.HTTPNotFound(reason="%s [%s] is not a master of node [%s]" % (str_master_type, name, node_name))
    db.commit()
    return web.Response(text="ok")



from click.testing import CliRunner
from joule.cli import main
from ..fake_joule import print_result_on_error
import unittest
import tempfile
import shutil
import os
from joule.errors import ApiError
from unittest.mock import patch
from joule.api import helpers, TcpNode

CAFILE = os.path.join(os.path.dirname(__file__), 'ca.joule.crt')

class TestNodeSetDefault(unittest.TestCase):

    def setUp(self):
        self.temp_dir = tempfile.TemporaryDirectory()
        os.environ["JOULE_USER_CONFIG_DIR"] = self.temp_dir.name
        self.ca_path = os.path.join(self.temp_dir.name, "ca.crt")
        shutil.copyfile(CAFILE, self.ca_path)
        
    def tearDown(self):
        self.temp_dir.cleanup()

    def test_sets_default_node(self):
        node1 = TcpNode("node1", "https://localhost:8088",
                        "node1_key")
        node2 = TcpNode("node2", "https://localhost:8089",
                        "node2_key")

        helpers.save_node(node1)
        helpers.save_node(node2)

        # set node2 as default
        runner = CliRunner()
        result = runner.invoke(main, ['node', 'default','node2'])
        print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        self.assertIn("node2", result.output)
        default_node = helpers.get_node()
        self.assertEqual(default_node.name, "node2")

        runner = CliRunner()
        result = runner.invoke(main, ['node', 'default','node1'])
        print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        self.assertIn("node1", result.output)
        default_node = helpers.get_node()
        self.assertEqual(default_node.name, "node1")

    @patch('joule.cli.node.default.set_default_node')
    def test_handles_api_error(self, set_default_node):
        set_default_node.side_effect = ApiError("test error")
        runner = CliRunner()
        result = runner.invoke(main, ['node', 'default','node2'])
        self.assertIn('test error', result.output)
        self.assertEqual(result.exit_code, 1)


import datetime

from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web
import asyncio
from sqlalchemy.orm import Session

from joule.models import folder, DataStream, Folder, Element, StreamInfo
import joule.controllers
from tests.controllers.helpers import create_db, MockStore
from joule import app_keys
from joule.constants import EndPoints

import testing.postgresql
psql_key = web.AppKey("psql", testing.postgresql.Postgresql)

class TestStreamController(AioHTTPTestCase):

    async def tearDownAsync(self):
        self.app[app_keys.db].close()
        self.app[psql_key].stop()
        await self.client.close()

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        # this takes a while, adjust the expected coroutine execution time
        loop = asyncio.get_running_loop()
        loop.slow_callback_duration = 2.0
        app[app_keys.db], app[psql_key] = create_db(["/folder1/stream1:float32[x, y, z]",
                                            "/folder2/deeper/stream2:int16[val1, val2]"])
        app[app_keys.data_store] = MockStore()
        return app


    async def test_stream_info(self):
        db: Session = self.app[app_keys.db]
        my_stream: DataStream = db.query(DataStream).filter_by(name="stream1").one()
        store: MockStore = self.app[app_keys.data_store]
        mock_info = StreamInfo(start=0, end=100, rows=200)
        store.set_info(my_stream, mock_info)
        # can query by id
        resp = await self.client.request("GET", "/stream.json?id=%d" % my_stream.id)
        actual = await resp.json()
        expected = my_stream.to_json({my_stream.id: mock_info})
        self.assertEqual(actual, expected)
        # can query by path
        payload = {'path': "/folder1/stream1"}
        resp = await self.client.request("GET", EndPoints.stream, params=payload)
        actual = await resp.json()
        self.assertEqual(actual, expected)
        # adding no-info skips the data store query
        payload = {'path': "/folder1/stream1", 'no-info': ''}
        resp = await self.client.request("GET", EndPoints.stream, params=payload)
        actual = await resp.json()
        self.assertEqual(actual, my_stream.to_json({}))


    async def test_stream_move(self):
        db: Session = self.app[app_keys.db]
        # move stream1 into folder3
        payload = {
            "src_path": "/folder1/stream1",
            "dest_path": "/new/folder3"
        }
        resp = await self.client.put(EndPoints.stream_move, json=payload)
        self.assertEqual(resp.status, 200)
        folder3 = db.query(Folder).filter_by(name="folder3").one()
        folder3_created_at = folder3.updated_at
        folder1 = db.query(Folder).filter_by(name="folder1").one()
        folder1_created_at = folder1.updated_at
        # check the destination
        self.assertEqual(folder3.data_streams[0].name, "stream1")
        self.assertEqual(folder3.parent.name, "new")
        # check the source
        self.assertEqual(len(folder1.data_streams), 0)

        # move stream1 back to folder1
        payload = {
            "src_path": "/new/folder3/stream1",
            "dest_id": folder1.id
        }
        resp = await self.client.put(EndPoints.stream_move, json=payload)
        self.assertEqual(resp.status, 200)
        # check the destination
        self.assertEqual(folder1.data_streams[0].name, "stream1")
        # check the source
        self.assertEqual(len(folder3.data_streams), 0)
        # make sure both the destination and source are updated
        self.assertGreater(folder3.updated_at, folder3_created_at)
        self.assertGreater(folder1.updated_at, folder1_created_at)

    async def test_stream_create(self):
        db: Session = self.app[app_keys.db]
        new_stream = DataStream(name="test", datatype=DataStream.DATATYPE.FLOAT32,
                                updated_at=datetime.datetime.utcnow())
        new_stream.elements = [Element(name="e%d" % j, index=j,
                                       display_type=Element.DISPLAYTYPE.CONTINUOUS) for j in range(3)]
        payload = {
            "dest_path": "/deep/new folder",
            "stream": new_stream.to_json()
        }
        resp = await self.client.post(EndPoints.stream, json=payload)

        self.assertEqual(resp.status, 200)
        # check the stream was created correctly
        created_stream: DataStream = db.query(DataStream).filter_by(name="test").one()
        self.assertEqual(len(created_stream.elements), len(new_stream.elements))
        self.assertEqual(created_stream.folder.name, "new folder")

        # can create by dest_id as well
        folder1: Folder = db.query(Folder).filter_by(name="folder1").one()
        new_stream.name = "test2"
        payload = {
            "dest_id": folder1.id,
            "stream": new_stream.to_json()
        }
        resp = await self.client.post(EndPoints.stream, json=payload)

        self.assertEqual(resp.status, 200)
        # check the stream was created correctly
        created_stream: DataStream = db.query(DataStream).filter_by(name="test2").one()
        self.assertEqual(len(created_stream.elements), len(new_stream.elements))
        self.assertEqual(created_stream.folder.name, "folder1")


    async def test_stream_delete_by_path(self):
        db: Session = self.app[app_keys.db]
        my_stream: DataStream = db.query(DataStream).filter_by(name="stream1").one()
        folder = my_stream.folder
        folder_created_at = folder.updated_at
        store: MockStore = self.app[app_keys.data_store]
        payload = {'path': "/folder1/stream1"}
        resp = await self.client.delete(EndPoints.stream, params=payload)
        self.assertEqual(resp.status, 200)
        # make sure the parent was updated
        self.assertGreater(folder.updated_at, folder_created_at)
        # make sure it was removed from the data store
        self.assertEqual(store.destroyed_stream_id, my_stream.id)
        # and the metadata
        self.assertEqual(0, db.query(DataStream).filter_by(name="stream1").count())


    async def test_stream_delete_by_id(self):
        db: Session = self.app[app_keys.db]
        my_stream: DataStream = db.query(DataStream).filter_by(name="stream1").one()
        store: MockStore = self.app[app_keys.data_store]
        payload = {'id': my_stream.id}
        resp = await self.client.delete(EndPoints.stream, params=payload)
        self.assertEqual(resp.status, 200)
        # make sure it was removed from the data store
        self.assertEqual(store.destroyed_stream_id, my_stream.id)
        # and the metadata
        self.assertEqual(0, db.query(DataStream).filter_by(name="stream1").count())


    async def test_stream_update(self):
        db: Session = self.app[app_keys.db]
        my_stream: DataStream = db.query(DataStream).filter_by(name="stream1").one()
        created_at = my_stream.updated_at
        folder_created_at = my_stream.folder.updated_at
        # change the stream name
        payload = {
            "id": my_stream.id,
            "stream": {"name": "new name"}
        }
        resp = await self.client.put(EndPoints.stream, json=payload)
        self.assertEqual(resp.status, 200)
        my_stream: DataStream = db.get(DataStream, my_stream.id)
        self.assertEqual("new name", my_stream.name)
        # make sure updated timestamps are more recent than created timestamps
        self.assertGreater(my_stream.updated_at, created_at)
        self.assertGreater(my_stream.folder.updated_at, folder_created_at)


import click
from joule.cli.lazy_group import LazyGroup

@click.group(name="follower",
             cls=LazyGroup,
             lazy_subcommands={"list": "joule.cli.follower.list.cli_list",
                               "delete": "joule.cli.follower.delete.follower_delete"})
def follower():
    """Manage node followers."""
    pass  




from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web
from sqlalchemy.orm import Session
import asyncio
from joule.models import Follower
import joule.controllers
from joule.api.node.node_info import NodeInfo
from tests.controllers.helpers import create_db, MockStore
from joule import app_keys
from joule.constants import EndPoints
from unittest import mock
import testing.postgresql
psql_key = web.AppKey("psql", testing.postgresql.Postgresql)

class TestFollowerController(AioHTTPTestCase):

    async def tearDownAsync(self):
        self.app[app_keys.db].close()
        await asyncio.to_thread(self.app[psql_key].stop)
        await self.client.close()

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        app[app_keys.cafile] = "stubbed_cafile"
        app[app_keys.db], app[psql_key] = await asyncio.to_thread(lambda: create_db([]))
        app[app_keys.name] = "test-node"
        return app


    async def test_lists_follower_nodes(self):
        db = self.app[app_keys.db]
        follower1 = Follower(name="test-follower",key="key1",location="http://othernode.local:8080")
        follower2 = Follower(name="test-follower2",key="key2",location="http://othernode2.local:8080")
        db.add(follower1)
        db.add(follower2)
        self.assertEqual(db.query(Follower).count(),2)
        resp = await self.client.get(EndPoints.followers)
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        self.assertEqual(len(data),2)
        self.assertIn(follower1.to_json(),data)
        self.assertIn(follower2.to_json(),data)

    @mock.patch('joule.controllers.follower_controller.TcpNode', )
    async def test_adds_follower_node(self, mock_node_class):
        mock_node = mock.AsyncMock()
        mock_node.info.return_value = NodeInfo(version="1.0",name="test-node-follower",path="/path",
                                               size_reserved=0,size_other=0,size_free=0,size_db=0)
        mock_node_class.return_value = mock_node
        # patch the TcpNode class to return a mock object
        resp = await self.client.post(EndPoints.follower, 
                                      json={"key":"key",
                                            "port":8080,
                                            "scheme":"http",
                                            "base_uri":"/endpoint"})
        # Follower's request is successful and gets *this* node's name
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        self.assertEqual(data['name'],"test-node")
        # Follower is added to the database
        db = self.app[app_keys.db]
        self.assertEqual(db.query(Follower).count(),1)
        follower = db.query(Follower).one()
        self.assertEqual(follower.name,"test-node-follower")
        self.assertEqual(follower.key,"key")
        # follower's IP address is based on request.remote which is just localhost
        self.assertEqual(follower.location,"http://127.0.0.1:8080/endpoint")
        # drop the follower from the database
        db.query(Follower).delete()

        # Follower can specify it's own name which will be used in place of the 
        # request.remote IP address
        resp = await self.client.post(EndPoints.follower, 
                                      json={"key":"key",
                                            "port":8080,
                                            "scheme":"https",
                                            "base_uri":"",
                                            "name": "follower-node",
                                            "name_is_host":True})
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        self.assertEqual(data['name'],"test-node")

        db = self.app[app_keys.db]
        self.assertEqual(db.query(Follower).count(),1)
        follower = db.query(Follower).one()
        self.assertEqual(follower.name,"test-node-follower")
        self.assertEqual(follower.key,"key")
        self.assertEqual(follower.location,"https://follower-node:8080")

    async def test_deletes_follower_nodes(self):
        db = self.app[app_keys.db]
        db.add(Follower(name="test-follower",key="key",location="http://othernode.local:8080"))
        self.assertEqual(db.query(Follower).count(),1)
        resp = await self.client.delete(EndPoints.follower,params={"name":"test-follower"})
        self.assertEqual(resp.status, 200)
        # the follower should be removed
        self.assertEqual(db.query(Follower).count(),0)



#!/usr/bin/python3
import time
import numpy as np
import asyncio
from joule import api
import sys


def main():
    asyncio.run(_run())


async def _run():
    node = api.get_node()
    await check_modules(node)
    await check_logs(node)
    await check_data(node)
    await node.close()


async def check_modules(node: api.BaseNode):
    """
    Test: check module status
    Goal:
      normal1: running some nonzero memory
      normal2: running some nonzero memory
      cadder:  running some nonzero memory
      broken:  present (could be running or failed)
      misconfigured: not present b/c can't be loaded
    """
    modules = await node.module_list(statistics=True)
    assert (len(modules) == 4)  # normal1,normal2,filter,broken
    for module in modules:
        if module.name in ['Normal1', 'Normal2', 'CAdder']:
            if module.statistics.pid is None:
                logs = await node.module_logs(module.name)
                print(module.name)
                for line in logs:
                    print(line)

            assert module.statistics.pid is not None, module.name
        elif module.name == 'Broken':
            pass
        else:
            assert 0  # unexpected module in status report


async def check_data(node: api.BaseNode):
    """
    Test: check data inserted correctly
    Goal:
      /bc/normal1/data is int32_1, has 1 interval with >500 samples
      /bc/normal2/subpath/data  is int32_1, >1 intervals each with 100 samples
      /bc/broken/data is float64_1, has separated intervals of data
      /bc/filter3/data is float64_1, has separated intervals of data
      both normal1 and normal2 have decimations
    """
    normal1_path = "/bc/normal1/data"
    normal2_path = "/bc/normal2/subpath/data"
    broken_path = "/bc/broken/data"
    filter1_path = "/bc/filter1/data"
    filter2_path = "/bc/filter2/data"
    for path in [normal1_path, normal2_path,
                 filter1_path, filter2_path,
                 broken_path]:
        # read 50 *live* samples from each stream
        pipe = await node.data_read(path)
        if 'broken' not in path:
            assert pipe.layout == "int32_1"
        else:
            assert pipe.layout == "float64_1"
        len_data = 0
        while await pipe.not_empty():
            data = await pipe.read()
            len_data += len(data)
            if len_data > 50:
                break
            pipe.consume(len(data))
        await pipe.close()

    # check intervals
    for path in [normal1_path, normal2_path,
                 filter1_path, filter2_path]:
        intervals = await node.data_intervals(path)
        assert len(intervals) == 1, 'path [%s] intervals %r' % (
            path, intervals)
    intervals = await node.data_intervals(broken_path)
    assert len(intervals) > 1, 'path [%s] intervals %r' % (
        broken_path, intervals)

    # read historic data to check if is correct

    # verify the filter module executed correctly
    # use the filter time bounds because to extract
    # the data to compare
    stream_info = await node.data_stream_info(filter1_path)
    p = await node.data_read(normal1_path,
                             start=stream_info.start,
                             end=stream_info.end,
                             max_rows=50)
    expected_data = await p.read(True)
    assert 0 < len(expected_data) <= 50
    expected_data[:, 1:] += 10.0
    await p.close()
    p = await node.data_read(filter1_path,
                             start=stream_info.start,
                             end=stream_info.end,
                             max_rows=50)
    actual_data = await p.read(True)
    verify_len = min(len(actual_data), len(expected_data))
    np.testing.assert_almost_equal(actual_data[:verify_len, :],
                                   expected_data[:verify_len, :])

    stream_info = await node.data_stream_info(filter2_path)
    p = await node.data_read(normal2_path,
                             start=stream_info.start,
                             end=stream_info.end,
                             max_rows=100)
    expected_data = await p.read(True)
    assert 0 < len(expected_data) <= 100
    expected_data[:, 1:] += 11.0
    await p.close()
    p = await node.data_read(filter2_path,
                             start=stream_info.start,
                             end=stream_info.end,
                             max_rows=100)
    actual_data = await p.read(True)
    verify_len = min(len(actual_data), len(expected_data))
    np.testing.assert_almost_equal(actual_data[:verify_len, :],
                                   expected_data[:verify_len, :])


async def check_logs(node: api.BaseNode):
    """
    Test: logs should contain info and stderr from modules
    Goal: 
      Normal1: says "starting" somewhere once
      Normal2: says "starting" somewhere once
      Filter: says "starting" somewhere once
      Broken: says "starting" and more than one "restarting"
      Misconfigured: no logs
    """
    for module_name in ["Normal1", "Normal2", "CAdder", "Broken"]:
        logs = await node.module_logs(module_name)
        num_starts = 0
        for line in logs:
            if "---starting module---" in line:
                num_starts += 1
        if module_name != 'Broken':
            assert num_starts == 1
        else:
            assert num_starts > 1


if __name__ == "__main__":
    main()
    print("OK")



import os
import argparse
import asyncio
import json
import unittest
import numpy as np

from unittest import mock
from joule.api import BaseNode

from joule.client import FilterModule
from joule.models import DataStream, Element, pipes
from tests import helpers
import warnings

warnings.simplefilter('always')


class SimpleFilter(FilterModule):
    # multiplies data by 2
    async def run(self, parsed_args, inputs, outputs):
        while await inputs['to_filter'].not_empty():
            data = await inputs['to_filter'].read()
            inputs['to_filter'].consume(len(data))
            data['data'] *= 2
            await outputs['from_filter'].write(data)

class TestFilterModule(helpers.AsyncTestCase):

    def setUp(self):
        super().setUp()
        # module output is a float32_3 stream
        self.output = DataStream(name="output", datatype=DataStream.DATATYPE.FLOAT32,
                                 elements=[Element(name="e%d" % j, index=j,
                                               display_type=Element.DISPLAYTYPE.CONTINUOUS) for j in range(3)])
        # module input is a float32_3 stream
        self.input = DataStream(name="input", datatype=DataStream.DATATYPE.FLOAT32,
                                elements=[Element(name="e%d" % j, index=j,
                                              display_type=Element.DISPLAYTYPE.CONTINUOUS) for j in range(3)])

    def test_writes_to_pipes(self):
        module = SimpleFilter()
        (r, w_module) = os.pipe()

        rf = pipes.reader_factory(r)
        from_filter = pipes.InputPipe(name="from_filter", stream=self.output, reader_factory=rf)
        (r_module, w) = os.pipe()

        wf = pipes.writer_factory(w)
        to_filter = pipes.OutputPipe(name="to_filter", stream=self.input, writer_factory=wf)

        pipe_arg = json.dumps(
            json.dumps({"outputs": {'from_filter': {'fd': w_module, 'id': 2, 'layout': self.output.layout}},
                        "inputs": {'to_filter': {'fd': r_module, 'id': 3, 'layout': self.input.layout}}}))
        data = helpers.create_data(self.input.layout)
        async def runner():
            await to_filter.write(data)
            await to_filter.close()
        asyncio.run(runner())
        args = argparse.Namespace(pipes=pipe_arg, socket="unset",
                                  node="", api_socket="", live=False,
                                  url='http://localhost:8080')
        # run the reader module

        class MockNode(BaseNode):
            def __init__(self):
                self.session = mock.Mock()
                self.session.close = unittest.mock.AsyncMock()
                pass

        with mock.patch('joule.client.base_module.node') as mock_node_pkg:
            node = MockNode()
            node.data_stream_get = unittest.mock.AsyncMock(return_value=self.output)
            mock_node_pkg.UnixNode = mock.Mock(return_value=node)
            module.start(args)
            # make sure the API was used to retrieve stream objects
            self.assertEqual(node.data_stream_get.await_count, 2)

        # check the output
        received_data = asyncio.run(from_filter.read())
        np.testing.assert_array_equal(data['timestamp'], received_data['timestamp'])
        np.testing.assert_array_almost_equal(data['data'] * 2, received_data['data'])
        asyncio.run(from_filter.close())

    def test_error_on_invalid_params(self):
        # if pipes is not set, must specify a module_config
        module = SimpleFilter()
        args = argparse.Namespace(socket='none', pipes='unset',
                                  url='http://localhost:8080',
                                  module_config='unset',
                                  node="", api_socket="",
                                  start_time=None, end_time=None)
        # run the module
        with self.assertLogs(level="ERROR") as logs:
            module.start(args)
        log_dump = ' '.join(logs.output).lower()
        self.assertTrue('module_config' in log_dump)
